import time
from datetime import datetime
import multiprocessing as mp
import os
from . import fortran_calls, readme_generators,cores, misc, magfield_inputs
import pandas as pd
import sys
import queue
import numpy as np

def OTSO_magfield(Locations,
           serverdata,livedata,vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,year,
           month,day,hour,minute,second,internalmag,externalmag,
           coordsystemIN,g,h,corenum,MHDfile,MHDcoordsys):

    magfieldInputArray = magfield_inputs.MagFieldInputs(Locations,
           serverdata,livedata,vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,year,
           month,day,hour,minute,second,internalmag,externalmag,
           coordsystemIN,g,h,corenum,MHDfile,MHDcoordsys)

    Locations = magfieldInputArray[0]
    DateArray = magfieldInputArray[1]
    Model = magfieldInputArray[2]
    IOPT = magfieldInputArray[3]
    WindArray = magfieldInputArray[4]
    CoordinateSystem = magfieldInputArray[5]
    Kp = magfieldInputArray[6]
    CoreNum = magfieldInputArray[7]
    LiveData = magfieldInputArray[8]
    serverdata = magfieldInputArray[9]
    g = magfieldInputArray[10]
    h = magfieldInputArray[11]

    ChildProcesses = []
    results = []

    LocationsList = np.array_split(Locations, CoreNum)

    start = time.time()
    print("OTSO Magfield Computation Started")
    sys.stdout.write(f"\r{0:.2f}% complete")


    try:
        if not mp.get_start_method(allow_none=True):
            mp.set_start_method('spawn')
    except RuntimeError:

        pass
# Create a shared message queue for the processes to produce/consume data
    ProcessQueue = mp.Manager().Queue()
    for Data in LocationsList:
        Child = mp.Process(target=fortran_calls.fortrancallMagfield,  args=(Data, DateArray, Model, IOPT, WindArray, CoordinateSystem, ProcessQueue,g,h,
                                                                            MHDfile, MHDcoordsys))
        ChildProcesses.append(Child)

    for a in ChildProcesses:
        a.start()

# Wait for child processes to complete

    results = []
    total_stations = len(Locations)
    processed = 0

    while processed < total_stations:
      try:
        # Check if the ProcessQueue has any new results
        result_df = ProcessQueue.get(timeout=0.001)  # Use timeout to avoid blocking forever
        results.append(result_df)
        processed += 1

        # Calculate and print the progress
        percent_complete = (processed / total_stations) * 100
        sys.stdout.write(f"\r{percent_complete:.2f}% complete")
        sys.stdout.flush()

      except queue.Empty:
        # Queue is empty, but processes are still running, so we continue checking
        pass
      
      time.sleep(0.0001)

    # Ensure that all processes have completed
    for b in ChildProcesses:
        b.join()

    combined_df = pd.concat(results, ignore_index=True)
    sorted_df = combined_df.sort_values(by=combined_df.columns[:3].tolist())

    print("\nOTSO Magfield Computation Complete")
    stop = time.time()
    Printtime = round((stop-start),3)
    print("Whole Program Took: " + str(Printtime) + " seconds")
    
    EventDate = datetime(year,month,day,hour,minute,second)
    README = readme_generators.READMEMagfield(EventDate, Model, IOPT, WindArray,
                                          CoordinateSystem, Printtime, LiveData, serverdata, kp)

    if LiveData == 1:
        misc.remove_files()
    
    return [sorted_df,README]


import time
from datetime import datetime
import multiprocessing as mp
import os
from . import fortran_calls, readme_generators,cores, misc, trace_inputs
import pandas as pd
import sys
import queue
import random
import numpy as np


def OTSO_trace(startaltitude,Coordsys,
           serverdata,livedata,vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,year,
           month,day,hour,minute,second,internalmag,externalmag,
           gyropercent,magnetopause,corenum,
           latstep,longstep,maxlat,minlat,maxlong,minlong,g,h,MHDfile,MHDcoordsys,spheresize):

    Anum = 1
    TraceInputArray = trace_inputs.TraceInputs(startaltitude,Coordsys,
           serverdata,livedata,vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,year,
           month,day,hour,minute,second,internalmag,externalmag,
           gyropercent,magnetopause,corenum,
           latstep,longstep,maxlat,minlat,maxlong,minlong,g,h,MHDfile,MHDcoordsys)

    LongitudeList = TraceInputArray[0]
    LatitudeList = TraceInputArray[1]
    Rigidity = TraceInputArray[2]
    ParticleArray = TraceInputArray[3]
    DateArray = TraceInputArray[4]
    Model = TraceInputArray[5]
    IntModel = TraceInputArray[6]
    IOPT = TraceInputArray[7]
    WindArray = TraceInputArray[8]
    Magnetopause = TraceInputArray[9]
    MaxStepPercent = TraceInputArray[10]/100
    EndParams = TraceInputArray[11]
    Zenith = TraceInputArray[12]
    Azimuth = TraceInputArray[13]
    CoreNum = TraceInputArray[14]
    Alt = TraceInputArray[15]
    LiveData = TraceInputArray[16]
    AntiCheck = TraceInputArray[17]
    g = TraceInputArray[18]
    h = TraceInputArray[19]

    ChildProcesses = []

    totalprocesses = len(LongitudeList)*len(LatitudeList)

    combined_coordinates = [(lat, lon) for lat in LatitudeList for lon in LongitudeList]

    NewCoreNum = misc.CheckCoreNumPlanet(CoreNum)
    FileNamesPlanet = []

    combined_coordinates = [(lat, lon) for lat in LatitudeList for lon in LongitudeList]
    for list in combined_coordinates:
        FileNamesPlanet.append(str(list[0]) + "_" + str(list[1]))
    DataPlanet = []
    i = 1
    for point,name in zip(combined_coordinates, FileNamesPlanet):
        Core = "Core " + str(i)
        DataPlanet.append([name,point[0],point[1],Alt,Zenith,Azimuth,Core])
        i = i + 1

    shuffled_list = DataPlanet.copy()
    random.shuffle(shuffled_list)
    DataLists = np.array_split(shuffled_list, CoreNum)
    CoreList = np.arange(1, CoreNum + 1)
    start = time.time()

    print("OTSO Trace Computation Started")
    sys.stdout.write(f"\r{0:.2f}% complete")
# Set the process creation method to 'forkserver'
    try:
        if not mp.get_start_method(allow_none=True):
            mp.set_start_method('spawn')
    except RuntimeError:

        pass

# Create a shared message queue for the processes to produce/consume data
    ProcessQueue = mp.Manager().Queue()
    for Data,Core in zip(DataLists, CoreList):
            Child = mp.Process(target=fortran_calls.fortrancallTrace,  args=(Data, Rigidity, DateArray, Model, IntModel, 
                                                                              ParticleArray, IOPT, WindArray, 
                                                                              Magnetopause, Coordsys, MaxStepPercent, EndParams,
                                                                              ProcessQueue,g,h, MHDfile, MHDcoordsys,
                                                                              spheresize))
            ChildProcesses.append(Child)

    for a in ChildProcesses:
        a.start()

    results = {}
    processed = 0
    while processed < totalprocesses:
        try:
            # Collect all results available in the queue at this moment
            result_collector = []
            while True:
                try:
                    result_df = ProcessQueue.get_nowait()  # Non-blocking, does not wait
                    result_collector.append(result_df)
                    processed += 1
                except queue.Empty:
                    break
    
            # Append all collected results to the main results list
            for x in result_collector:
                results.update(x)
    
            # Calculate and print the progress
            percent_complete = (processed / totalprocesses) * 100
            sys.stdout.write(f"\r{percent_complete:.2f}% complete")
            sys.stdout.flush()
    
        except queue.Empty:
            # Queue is empty, but processes are still running, so we continue checking
            pass
        
        # Wait for 5 seconds before the next iteration
        time.sleep(0.0001)

    for b in ChildProcesses:
        b.join()

    print("\nOTSO Trace Computation Complete")

    stop = time.time()
    Printtime = round((stop-start),3)
    print("Whole Program Took: " + str(Printtime) + " seconds")

    # Sorting the dictionary by its keys
    sorted_results = dict(sorted(results.items(), key=lambda item: parse_key(item[0])))
    
    EventDate = datetime(year,month,day,hour,minute,second)
    readme = readme_generators.READMETrace(Data, EventDate, Model, IntModel, 
                                             AntiCheck, IOPT, WindArray, Magnetopause, Printtime,
                                             maxlat,maxlong,minlat,minlong, latstep, longstep,
                                             MaxStepPercent*100, EndParams, 
                                             LiveData, serverdata, kp)

    if LiveData == 1:
        misc.remove_files()

    return [sorted_results,readme]


def parse_key(key):
    lat, lon = key.split('_')
    return (int(lat), int(lon))



import numpy as np
from datetime import datetime,timedelta
import os
from . import date, solar_wind, stations
from . import misc, Request, Server

def FlightInputs(latitudes,longitudes,dates,altitudes,cutoff_comp,minaltitude,maxdistance,maxtime,
           serverdata,livedata,vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,Anum,anti,internalmag,externalmag,
           intmodel,startrigidity,endrigidity,rigiditystep,rigidityscan,
           coordsystem,gyropercent,magnetopause,corenum,azimuth,zenith,g,h,asymptotic,asymlevels,unit,
           MHDfile, MHDcoordsys):
    
    DateArrayList = []
    for x in dates:
         DateCreate = date.Date(x)
         DateArray = DateCreate.GetDate()
         DateArrayList.append(DateArray)

    if unit not in ["GeV", "GV"]:
         print("Please enter a valid asymlevel unit: ""GeV"" or ""GV"" ")
         exit()

    if asymptotic not in ["YES","NO"]:
     print("Please enter a valid asymptotic value: ""YES"" or ""NO"" ")
     exit()

    if anti == "YES":
         AntiCheck = 1
    elif anti == "NO":
         AntiCheck = 0
    else:
         print("Please enter a valid anti value: ""YES"" or ""NO"" ")
         exit()

    if magnetopause == "Sphere":
         Magnetopause = 0
    elif magnetopause == "aFormisano":
         Magnetopause = 1
    elif magnetopause == "Sibeck":
         Magnetopause = 2
    elif magnetopause == "Kobel":
         Magnetopause = 3
    elif magnetopause == "NONE":
         Magnetopause = 99
    else:
         print("Please enter a valid magnetopause model: ""Sphere"", ""aFormisano"", ""Sibeck"", ""Kobel"", ""NONE"" ")


    if intmodel == "4RK":
         IntModel = 1
    elif intmodel == "Boris":
         IntModel = 2
    elif intmodel == "Vay":
         IntModel = 3
    elif intmodel == "HC":
         IntModel = 4
    else:
         print("Please enter a valid intmodel model: ""4RK"", ""Boris"", ""Vay"", ""HC"" ")
         exit()

    if cutoff_comp == "Vertical":
         CutoffComputation = 0
         Zenith = 0
         Azimuth = 0
    elif cutoff_comp == "Apparent":
         CutoffComputation = 1
         Zenith = 0
         Azimuth = 0
    elif cutoff_comp == "Custom":
         CutoffComputation = 2
         Zenith = zenith
         Azimuth = azimuth
    else:
         print("Please enter a valid cutoff_comp: ""Vertical"", ""Apparent"", or ""Custom"" ")
         exit()

    if serverdata == "ON":
         ServerData = 1
    elif serverdata == "OFF":
         ServerData = 0
    else:
         print("Please enter a valid serverdata value: ""ON"" or ""OFF"" ")
         exit()

    if livedata == "ON":
         LiveData = 1
    elif livedata == "OFF":
         LiveData = 0
    else:
         print("Please enter a valid livedata value: ""ON"" or ""OFF"" ")
         exit()
    
    variablelist = [vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp]
    variablelist2 = [latitudes,longitudes,dates,vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp]
    
    if serverdata == "OFF" and livedata == "OFF":
         if any(not lst for lst in variablelist):
              print("If not using livedata or server data you must provide full lists for the input variables\n"
              "vx,vy,vz,by,bz,density,pdyn,Dst,G1,G2,G3,W1,W2,W3,W4,W5,W6,kp")
              exit()
         if any(len(lst) != len(variablelist2[0]) for lst in variablelist2):
              print("If not using livedata or server data all provided variable lists must be the same length")
              exit()
    
    if internalmag == "NONE":
         Internal = 0
         if not g or not h: 
            g = [0] * 105
            h = [0] * 105
    elif internalmag == "IGRF":
         Internal = 1
         if not g or not h: 
            g = [0] * 105
            h = [0] * 105
    elif internalmag == "Dipole":
         Internal = 2
         if not g or not h: 
            g = [0] * 105
            h = [0] * 105
    elif internalmag == "Custom Gauss":
         Internal = 4
         if not g or not h:
              print("Please enter values for the g and h Gaussian coefficients to use the Custom Gauss option")
              exit()
         elif len(g) != 105:
              print(f"There should be 105 g coefficents in the inputted list, you have entered {len(g)}")
              exit()
         elif len(h) != 105:
              print(f"There should be 105 h coefficents in the inputted list, you have enetered {len(h)}")
    else:
         print("Please enter a valid internalmag model: ""NONE"",""IGRF"",""Dipole"", or ""Custom Gauss""")
         exit()

         
    if externalmag == "NONE":
         External = 0
    elif externalmag == "TSY87short":
         External = 1
    elif externalmag == "TSY87long":
         External = 2
    elif externalmag == "TSY89":
         External = 3
    elif externalmag == "TSY96":
         External = 4
    elif externalmag == "TSY01":
         External = 5
    elif externalmag == "TSY01S":
         External = 6
    elif externalmag == "TSY04":
         External = 7
    elif externalmag == "TSY89_BOBERG":
         External = 8
    elif externalmag == "MHD":
         External = 99
         if not os.path.exists(MHDfile):
            print(f"The file '{MHDfile}' does not exist.")
            exit()
    else:
         print("Please enter a valid externalmag model: ""NONE"", ""TSY87short"",""TSy87long"",""TSY89"",""TSY89_BOBERG"",""TSY96"",""TSY01"",""TSY01S"",""TSY04""")
         exit()

    if coordsystem not in ["GDZ","GEO","GSM","GSE","SM","GEI","MAG","SPH","RLL"]:
         print("Please select a valid coordsystem: ""GDZ"", ""GEO"", ""GSM"", ""GSE"", ""SM"", ""GEI"", ""MAG"", ""SPH"", ""RLL""")
         exit()

    
    if rigidityscan == "ON":
         Rscan = 1
    elif rigidityscan == "OFF":
         Rscan = 0
    else:
         print("Please enter a valid rigidityscan value: ""ON"" or ""OFF"" ")
         exit()

    #IOPTinput = misc.IOPTprocess(kp[0])
    #KpS = 0

    KpList = []


    WindArrayList = []
    IOPTList = []
    i = 0
    for x in dates:
         
       if ServerData == 1:
          if int(x.year) >= 1981:
               Server.DownloadServerFile(int(x.year))
          elif int(x.year) < 1981 and int(x.year) > 1963:
               Server.DownloadServerFileLowRes(int(x.year))
          else:
               print("Server data only valid for 1963 to present, please enter a valid date.")
          ByS, BzS, VS, DensityS, PdynS, KpS, DstS, G1S, G2S, G3S, W1S, W2S, W3S, W4S, W5S, W6S = Server.GetServerData(x,External)
          KpList.append(KpS)
          IOPTinput = misc.IOPTprocess(KpS)
          IOPTList.append(IOPTinput)
          vytemp = 0
          vztemp = 0
          WindCreate = solar_wind.Solar_Wind(VS, vytemp, vztemp, ByS, BzS, DensityS, PdynS, DstS, G1S, G2S, G3S, W1S, W2S, W3S, W4S, W5S, W6S, KpS)
          WindArray = WindCreate.GetWind()
          WindArrayList.append(WindArray)
          
          if LiveData == 1:
               misc.DateCheck(x)
               DstLive, VxLive, DensityLive, ByLive, BzLive, IOPTLive, G1Live, G2Live, G3Live, KpLive = Request.Get_Data(x)
               PdynLive = misc.Pdyn_comp(DensityLive,VxLive)
               KpList.append(KpLive)
               IOPTinput = IOPTLive
               IOPTList.append(IOPTinput)
               vytemp = 0
               vztemp = 0
               WindCreate = solar_wind.Solar_Wind(VxLive, vytemp, vztemp, ByLive, BzLive, DensityLive, PdynLive, DstLive, G1Live, G2Live, G3Live, W1, W2, W3, W4, W5, W6, KpLive)
               WindArray = WindCreate.GetWind()
               WindArrayList.append(WindArray)

       if ServerData == 0 and LiveData == 0:
          if vx[i] > 0:
               vx[i] = -1*vx[i]
          WindCreate = solar_wind.Solar_Wind(vx[i], vy[i], vz[i], by[i], bz[i], density[i], pdyn[i], Dst[i], G1[i], G2[i], G3[i], W1[i], W2[i], W3[i], W4[i], W5[i], W6[i], kp[i])
          WindArray = WindCreate.GetWind()
          KpList.append(kp[i])
          IOPTinput = misc.IOPTprocess(kp[i])
          IOPTList.append(IOPTinput)
          WindArrayList.append(WindArray)
          i += 1

    RigidityArray = [startrigidity,endrigidity,rigiditystep]

    MagFieldModel = np.array([Internal,External])

    EndParams = [minaltitude,maxdistance,maxtime]
    
    stationslist = []
    for lat,long,alt in zip(latitudes,longitudes,altitudes):
         station = ["temp",lat,long,alt,Zenith,Azimuth]
         stationslist.append(station)
         misc.ParamCheck(alt,2000,Internal,EndParams)
         

    
    ParticleArray = [Anum,AntiCheck]

    FlightInputArray = [RigidityArray,DateArrayList,MagFieldModel,IntModel,ParticleArray,IOPTList,WindArrayList,Magnetopause,
    coordsystem,gyropercent,EndParams, stationslist, CutoffComputation, Rscan, KpList, corenum, LiveData, serverdata,g,h]

    return FlightInputArray


import requests
import pandas as pd
import time
import os
import json
from datetime import datetime, timedelta
from .Extract import *
from .Gvalues import *

output_directory = os.path.dirname(os.path.realpath(__file__)) 

def fetch_data_Space(urlSpace):
    try:
        response = requests.get(urlSpace)
        response.raise_for_status()  
        return response.json()  
    except requests.exceptions.RequestException as e:
        #print(f"Error fetching data: {e}")
        return None
    
def fetch_data_Mag(urlMag):
    try:
        response = requests.get(urlMag)
        response.raise_for_status()  
        return response.json()  
    except requests.exceptions.RequestException as e:
        #print(f"Error fetching data: {e}")
        return None

def fetch_data_Dst(urlDst):
    try:
        responseDst = requests.get(urlDst)
        responseDst.raise_for_status()  
        return responseDst.text  
    except requests.exceptions.RequestException as e:
        #print(f"Error fetching data: {e}")
        return None
    
def fetch_data_Kp(urlKp):
    try:
        responseKp = requests.get(urlKp)
        responseKp.raise_for_status()  
        return responseKp.text  
    except requests.exceptions.RequestException as e:
        #print(f"Error fetching data: {e}")
        return None

def get_next_version():
    existing_files = os.listdir(output_directory)
    version_numbers = []

    for file_name in existing_files:
        if file_name.startswith("space_data_") and (file_name.endswith(".csv") or file_name.endswith(".json")):
            try:
                version = int(file_name.split('_')[-1].split('.')[0])
                version_numbers.append(version)
            except ValueError:
                continue
    
    if version_numbers:
        return max(version_numbers) + 1
    else:
        return 1

def save_to_json_space(data, version):
    json_file = os.path.join(output_directory, f'space_data.json')
    with open(json_file, 'w') as file:
        json.dump(data, file, indent=4)

def save_to_json_magnetic(data, version):
    json_file = os.path.join(output_directory, f'Magnetic_data.json')
    with open(json_file, 'w') as file:
        json.dump(data, file, indent=4)

def save_to_txt_Dst(data, version):
    Dst_file = os.path.join(output_directory, f'Dst_data.txt')
    with open(Dst_file, "w") as fileDst:
            fileDst.write(data)

def save_to_txt_Kp(data, version):
    Kp_file = os.path.join(output_directory, f'Kp_data.txt')
    with open(Kp_file, "w") as fileKp:
            fileKp.write(data)

def save_to_csv_Space(data, version):
    csv_file = os.path.join(output_directory, f'space_data.csv')
    try:
        if isinstance(data, list) and all(isinstance(row, list) for row in data):
            headers = data[0]
            rows = data[1:]

            df = pd.DataFrame(rows, columns=headers)
            df['fetch_timestamp'] = pd.Timestamp.now()
            df.to_csv(csv_file, index=False)

    except Exception as e:
        print(f"Error converting JSON to CSV: {e}")

def save_to_csv_Mag(data, version):
    csv_file = os.path.join(output_directory, f'Magnetic_data.csv')
    try:
        if isinstance(data, list) and all(isinstance(row, list) for row in data):
            headers = data[0]
            rows = data[1:]
            
            df = pd.DataFrame(rows, columns=headers)
            df['fetch_timestamp'] = pd.Timestamp.now()
            df.to_csv(csv_file, index=False)

    except Exception as e:
        print(f"Error converting JSON to CSV: {e}")

def Get_Data(current_time):

    os.makedirs(output_directory, exist_ok=True)

    #current_time = datetime.utcnow()
    #current_time = current_time - timedelta(hours=3)
    current_month = f"{current_time.month:02d}"
    current_year = current_time.year
    curretnt_year_two_digits = current_year % 100

    dststring1 = str(current_year)+str(current_month)
    dststring2 = str(curretnt_year_two_digits)+str(current_month)
    
    urlSpace = 'https://services.swpc.noaa.gov/products/solar-wind/plasma-7-day.json'
    urlMag = 'https://services.swpc.noaa.gov/products/solar-wind/mag-7-day.json'
    urlDst = 'https://wdc.kugi.kyoto-u.ac.jp/dst_realtime/'+dststring1+'/dst'+dststring2+'.for.request'
    urlKp = 'https://kp.gfz-potsdam.de/app/files/Kp_ap_nowcast.txt'

    data_Space = fetch_data_Space(urlSpace)
    data_Mag = fetch_data_Mag(urlMag)
    data_Dst = fetch_data_Dst(urlDst)
    data_Kp = fetch_data_Kp(urlKp)
    if data_Space:
        version = get_next_version()
        save_to_json_space(data_Space, version)
        save_to_json_magnetic(data_Space, version)
        save_to_csv_Space(data_Space, version)
        save_to_csv_Mag(data_Mag, version)
        save_to_txt_Dst(data_Dst, version)
        save_to_txt_Kp(data_Kp, version)
    else:
        print("Failed to obtain online data")
    
    DstFile = os.path.join(output_directory, f'Dst_data.txt')
    Dst,daily_dst = extract_dst_value(DstFile, current_time)
    Speed,Density = extract_solar_wind(current_time)
    By, Bz = extract_magnetic(current_time)
    Kp,IOPT = extract_kp_index(current_time)
    G1,G2,G3 = TSY01_Constants(By,Bz,-1*Speed,Density)
    Speed = round(Speed,3)
    Density = round(Density,3)
    By = round(By,3)
    Bz = round(Bz,3)
    G1 = round(G1,3)
    G2 = round(G2,3)
    G3 = round(G3,3)

    return Dst, Speed, Density, By, Bz, IOPT, G1, G2, G3, Kp



import os
import sys


def _ensure_setup_complete():
    from .Callibrate import setup as setup_func
    package_dir = os.path.dirname(os.path.abspath(__file__))
    setup_marker = os.path.join(package_dir, 'setup_complete.txt')
    if not os.path.exists(setup_marker):
        setup_func()
    return

def cutoff(*args, **kwargs):
    _ensure_setup_complete()
    from .Cutoff import cutoff as cutoff_func
    return cutoff_func(*args, **kwargs)

def cone(*args, **kwargs):
    _ensure_setup_complete()
    from .Cone import cone as cone_func
    return cone_func(*args, **kwargs)

def planet(*args, **kwargs):
    _ensure_setup_complete()
    from .Planet import planet as planet_func
    return planet_func(*args, **kwargs)

def trajectory(*args, **kwargs):
    _ensure_setup_complete()
    from .Trajectory import trajectory as trajectory_func
    return trajectory_func(*args, **kwargs)

def flight(*args, **kwargs):
    _ensure_setup_complete()
    from .Flight import flight as flight_func
    return flight_func(*args, **kwargs)

def magfield(*args, **kwargs):
    _ensure_setup_complete()
    from .Magfield import magfield as magfield_func
    return magfield_func(*args, **kwargs)

def coordtrans(*args, **kwargs):
    _ensure_setup_complete()
    from .Coordtrans import coordtrans as coordtrans_func
    return coordtrans_func(*args, **kwargs)

def trace(*args, **kwargs):
    _ensure_setup_complete()
    from .Trace import trace as trace_func
    return trace_func(*args, **kwargs)

def clean(*args, **kwargs):
    from .OTSO import clean as clean_func
    return clean_func(*args, **kwargs)

def addstation(*args, **kwargs):
    from .OTSO import AddStation as addstation_func
    Name, Latitude, Longitude = sys.argv[1], float(sys.argv[2]), float(sys.argv[3])
    addstation_func(Name, Latitude, Longitude)
    return

def removestation(*args, **kwargs):
    from .OTSO import RemoveStation as removestation_func
    Name = sys.argv[1]
    removestation_func(Name)
    return

def liststations(*args, **kwargs):
    from .OTSO import ListStations as liststations_func
    return liststations_func(*args, **kwargs)


import time
from datetime import datetime
import multiprocessing as mp
import os
from . import fortran_calls, readme_generators, misc, date
import pandas as pd
import sys
import queue
import numpy as np

def OTSO_coordtrans(Locations,Dates,CoordIN,CoordOUT,corenum):
    
    if CoordIN not in ["GDZ","GEO","GSM","GSE","SM","GEI","MAG","SPH","RLL"]:
         print("Please select a valid CoordIN: ""GDZ"", ""GEO"", ""GSM"", ""GSE"", ""SM"", ""GEI"", ""MAG"", ""SPH"", ""RLL""")
         exit()
    if CoordOUT not in ["GDZ","GEO","GSM","GSE","SM","GEI","MAG","SPH","RLL"]:
         print("Please select a valid CoordOUT: ""GDZ"", ""GEO"", ""GSM"", ""GSE"", ""SM"", ""GEI"", ""MAG"", ""SPH"", ""RLL""")
         exit()

    ChildProcesses = []
    results = []
    DateArrayList = []

    for x in Dates:
          DateCreate = date.Date(x)
          DateArray = DateCreate.GetDate()
          DateArrayList.append(DateArray)

    LocationsList = np.array_split(Locations, corenum)
    DateArrayList = np.array_split(DateArrayList, corenum)


    start = time.time()
    print("OTSO Coordtrans Computation Started")
    sys.stdout.write(f"\r{0:.2f}% complete")


    try:
        if not mp.get_start_method(allow_none=True):
            mp.set_start_method('spawn')
    except RuntimeError:

        pass
# Create a shared message queue for the processes to produce/consume data
    ProcessQueue = mp.Manager().Queue()
    for Data,Date in zip(LocationsList,DateArrayList):
        Child = mp.Process(target=fortran_calls.fortrancallCoordtrans,  args=(Data, Date, CoordIN, CoordOUT, ProcessQueue))
        ChildProcesses.append(Child)

    for a in ChildProcesses:
        a.start()

# Wait for child processes to complete

    results = []
    total_stations = len(Locations)
    processed = 0

    while processed < total_stations:
      try:
        # Check if the ProcessQueue has any new results
        result_df = ProcessQueue.get(timeout=0.001)  # Use timeout to avoid blocking forever
        results.append(result_df)
        processed += 1

        # Calculate and print the progress
        percent_complete = (processed / total_stations) * 100
        sys.stdout.write(f"\r{percent_complete:.2f}% complete")
        sys.stdout.flush()

      except queue.Empty:
        # Queue is empty, but processes are still running, so we continue checking
        pass
      
      time.sleep(0.0001)

    # Ensure that all processes have completed
    for b in ChildProcesses:
        b.join()

    combined_df = pd.concat(results, ignore_index=True)
    sorted_df = combined_df.sort_values(by=combined_df.columns[:4].tolist())

    print("\nOTSO Coordtrans Computation Complete")
    stop = time.time()
    Printtime = round((stop-start),3)
    print("Whole Program Took: " + str(Printtime) + " seconds")
    
    README = readme_generators.READMECoordtrans(CoordIN,CoordOUT,Printtime)
    
    return [sorted_df, README]


import sys, os
import subprocess
import platform

def setup():
    #print("Running setup tasks...")
    a = DeleteLibs()
    #DarwinLib()
    Generatefile(a)
    #print("Setup complete")

def DarwinLib():
    if platform.system() == "darwin": 
        so_files = []
        script_dir = os.path.dirname(os.path.abspath(__file__))
        package_path = os.path.join(script_dir, 'Parameters', 'functions')

        for root, _, files in os.walk(package_path):
            for file in files:
                if file.endswith(('.so', '.pyd')): 
                    so_files.append(os.path.join(root, file))

        for lib_path in so_files:
            try:
                subprocess.run(['install_name_tool', '-add_rpath', '@loader_path', package_path], check=True)
                print(f"Updated Rpath for {package_path}")
            except subprocess.CalledProcessError as e:
                print(f"Error updating Rpath for {package_path}: {e}")


def DeleteLibs():
    script_dir = os.path.dirname(os.path.abspath(__file__))
    package_path = os.path.join(script_dir, 'Parameters', 'functions')
    system_type = platform.system().lower()
    python_version = f"{sys.version_info.major}{sys.version_info.minor}"

    if system_type == "windows":
        system_type = "win"

    error_messages = []

    for filename in os.listdir(package_path):
        file_path = os.path.join(package_path, filename)
        if os.path.isfile(file_path) and filename.endswith(('.so', '.pyd')):
            if system_type not in filename or python_version not in filename:
                try:
                    os.remove(file_path)
                except Exception as e:
                    error_messages.append(f"Could not delete {file_path}: {e}")

    return "\n".join(error_messages) if error_messages else ""


def Generatefile(a):
    script_dir = os.path.dirname(os.path.abspath(__file__))
    marker_path = os.path.join(script_dir, 'setup_complete.txt')

    try:
        with open(marker_path, 'w') as f:
            f.write('Setup has been completed.\n')
            f.write(a)
        #print(f"Setup marker file created at: {marker_path}")
    except Exception as e:
        print(f"Error creating setup marker file: {e}")




from . import MiddleMan as OTSOLib
from . import OmniPull
import os
from datetime import datetime, timedelta
import pandas as pd
import numpy as np
import sys
import platform

def round_to_nearest_five_minutes(date):
    if date.minute % 5 == 0 and date.second == 0 and date.microsecond == 0:
        return date.replace(second=0, microsecond=0)
    return (date + timedelta(minutes=5 - date.minute % 5)).replace(second=0, microsecond=0)

def GetServerData(Date, External):
    OMNIYEAR = int(Date.year)
    RoundedDate = round_to_nearest_five_minutes(Date)
    Date, By, Bz, V, Density, Pdyn, Kp, Dst, G1, G2, G3, W1, W2, W3, W4, W5, W6 = ExtractServerData(RoundedDate, External)

    return By, Bz, V, Density, Pdyn, Kp, Dst, G1, G2, G3, W1, W2, W3, W4, W5, W6


def ExtractServerData(RoundedDate,External):
    year = RoundedDate.year
    source_file = f'{year}_TSY_Inputs.csv'
    TSY_folder = os.path.join(os.path.dirname(os.path.realpath(__file__)), "ServerData")
    TSY_file = os.path.join(TSY_folder, os.path.basename(source_file))

    if not os.path.exists(TSY_file):
        print(f"File for year {year} not found in the directory.")
        return

    try:
        df = pd.read_csv(TSY_file)
    except Exception as e:
        print(f"Error reading {source_file}: {e}")
        return

    if 'Date' not in df.columns:
        print(f"'Date' column not found in {source_file}")
        return

    df['Date'] = pd.to_datetime(df['Date'], errors='coerce')

    matching_row = df[df['Date'] == RoundedDate]

    #print(RoundedDate)

    if matching_row.empty:
        print(f"No matching row found for datetime: {RoundedDate}")
        return None
    else:
        row_data = matching_row.iloc[0]  

        Date = row_data['Date'] if 'Date' in row_data else None
        By = row_data['By'] if 'By' in row_data else None
        Bz = row_data['Bz'] if 'Bz' in row_data else None
        V = row_data['V'] if 'V' in row_data else None
        Density = row_data['Density'] if 'Density' in row_data else None
        Pdyn = row_data['Pdyn'] if 'Pdyn' in row_data else None
        Kp = row_data['Kp'] if 'Kp' in row_data else None
        Dst = row_data['Dst'] if 'Dst' in row_data else None
        G1 = row_data['G1'] if 'G1' in row_data else None
        G2 = row_data['G2'] if 'G2' in row_data else None
        G3 = row_data['G3'] if 'G3' in row_data else None
        W1 = row_data['W1'] if 'W1' in row_data else None
        W2 = row_data['W2'] if 'W2' in row_data else None
        W3 = row_data['W3'] if 'W3' in row_data else None
        W4 = row_data['W4'] if 'W4' in row_data else None
        W5 = row_data['W5'] if 'W5' in row_data else None
        W6 = row_data['W6'] if 'W6' in row_data else None

        if External == 7 and (np.isnan(W1) or W1 == 0) and (np.isnan(V) or V == 0) and (np.isnan(Bz) or Bz == 0):
           print("ERROR: No TSY04 parameters found for given time. \nOTSO program will now terminate.")
           exit()
        elif External == 6 and (np.isnan(G3) or G3 == 0) and (np.isnan(V) or V == 0) and (np.isnan(Bz) or Bz == 0):
           print("ERROR: No TSY01S parameters found for given time. \nOTSO program will now terminate.")
           exit()
        elif External == 5 and (np.isnan(G1) or G1 == 0) and (np.isnan(G2) or G2 == 0) and (np.isnan(V) or V == 0) and  (np.isnan(Bz) or Bz == 0):
           print("ERROR: No TSY01 parameters found for given time. \nOTSO program will now terminate.")
           exit()
        elif External == 4 and (np.isnan(V) or V == 0) and (np.isnan(Bz) or Bz == 0):
           print("ERROR: No TSY96 parameters found for given time. \nOTSO program will now terminate.")
           exit()

        if V > 0:
            V = -1*V

        return Date, By, Bz, V, Density, Pdyn, Kp, Dst, G1, G2, G3, W1, W2, W3, W4, W5, W6


def DownloadServerFile(OMNIYEAR):
    source_file = f'{OMNIYEAR}_TSY_Inputs.csv'
    TSY_folder = os.path.join(os.path.dirname(os.path.realpath(__file__)), "ServerData")
    TSY_file = os.path.join(TSY_folder, source_file)

    if not os.path.exists(TSY_file):
        print(f"Data for {OMNIYEAR} does not exist in OTSO files.")
        print(f'Attempting to download data for {OMNIYEAR}')
        OmniPull.PullOMNI(OMNIYEAR)
        DIRECTORY = os.path.join(os.path.dirname(os.path.dirname(__file__)),"functions")
        length = len(DIRECTORY)
        if platform.system() == "Windows":
            OTSOLib.gettsy04datawindows(OMNIYEAR, length)
            OmniPull.OMNI_to_csv(OMNIYEAR)
            OmniPull.TSY01(f'{OMNIYEAR}_TSY_Data.csv')
            OmniPull.TSY01(f'omni_{OMNIYEAR}_high_res.csv')
            OmniPull.TSY01(f'omni_{OMNIYEAR}_low_res.csv')
            OmniPull.Combine(f'{OMNIYEAR}_TSY_Data.csv', f'omni_{OMNIYEAR}_high_res.csv', f'omni_{OMNIYEAR}_low_res.csv', OMNIYEAR)
            OmniPull.Omnidelete(OMNIYEAR)
            print(f'Finished downloading data for {OMNIYEAR}.')
        elif platform.system() == "Linux"  or "Darwin":
            OTSOLib.gettsy04datalinux(OMNIYEAR, DIRECTORY, length)
            OmniPull.OMNI_to_csv(OMNIYEAR)
            OmniPull.TSY01(f'{OMNIYEAR}_TSY_Data.csv')
            OmniPull.TSY01(f'omni_{OMNIYEAR}_high_res.csv')
            OmniPull.TSY01(f'omni_{OMNIYEAR}_low_res.csv')
            OmniPull.Combine(f'{OMNIYEAR}_TSY_Data.csv', f'omni_{OMNIYEAR}_high_res.csv', f'omni_{OMNIYEAR}_low_res.csv', OMNIYEAR)
            OmniPull.Omnidelete(OMNIYEAR)
            print(f'Finished downloading data for {OMNIYEAR}.')

    return


def DownloadServerFileLowRes(OMNIYEAR):
    source_file = f'{OMNIYEAR}_TSY_Inputs.csv'
    TSY_folder = os.path.join(os.path.dirname(os.path.realpath(__file__)), "ServerData")
    TSY_file = os.path.join(TSY_folder, os.path.basename(source_file))
    FilePath = os.path.join(os.path.dirname(os.path.realpath(__file__)), f'omni_{OMNIYEAR}_low_res.csv')

    if not os.path.exists(TSY_file):
        print(f"Data for {OMNIYEAR} does not exist in OTSO files.")
        print(f'Attempting to download data for {OMNIYEAR}')
        OmniPull.PullOMNILowRes(OMNIYEAR)
        OmniPull.CombineLowRes(FilePath,OMNIYEAR)
        OmniPull.OmnideleteLowRes(OMNIYEAR)
        print(f'Finished downloading data for {OMNIYEAR}.')


    


def magfield(Locations,serverdata="OFF",livedata="OFF",vx=-500,vy=0,vz=0,by=5,bz=5,density=1,pdyn=0,Dst=0,
           G1=0,G2=0,G3=0,W1=0,W2=0,W3=0,W4=0,W5=0,W6=0,kp=0,year=2024,
           month=1,day=1,hour=12,minute=0,second=0,internalmag="IGRF",externalmag="TSY89",
           coordsystem="GEO", g=None,h=None,corenum=1,MHDfile=None,MHDcoordsys=None):
    from .Parameters.functions import otso_magfield
    import psutil

    if corenum is None:
       corenum = psutil.cpu_count(logical=False) - 2
       if corenum <= 0:
          corenum = 1
    
    arguments = locals()
    for arg in arguments:
       if arguments[arg] is None:
          arguments[arg] = []

    magfield = otso_magfield.OTSO_magfield(Locations,serverdata,livedata,vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,year,
           month,day,hour,minute,second,internalmag,externalmag,
           coordsystem,arguments["g"],arguments["h"],corenum,MHDfile,MHDcoordsys)
    
    return magfield


from . import MiddleMan as OTSOLib
import time
import os
import shutil
from datetime import datetime
import pandas as pd
import glob
import csv
import numpy as np
import multiprocessing as mp
from . import date
import gc
import sys
import psutil
import tracemalloc

def MHDinitialise(MHDfile):
      data = pd.read_csv(MHDfile)
      x1 = data["X"].values.astype(np.float64)
      y1 = data["Y"].values.astype(np.float64)
      z1 = data["Z"].values.astype(np.float64)
      bx = data["Bx"].values.astype(np.float64)
      by = data["By"].values.astype(np.float64)
      bz = data["Bz"].values.astype(np.float64)
  
      XU = np.unique(x1)
      YU = np.unique(y1)
      ZU = np.unique(z1)
      
      XUlen, YUlen, ZUlen = len(XU), len(YU), len(ZU)
  
      ix = np.searchsorted(XU, x1) - 1
      iy = np.searchsorted(YU, y1) - 1
      iz = np.searchsorted(ZU, z1) - 1
  
      MHDposition = np.zeros((XUlen, YUlen, ZUlen, 3), dtype=np.float64)
      MHDB = np.zeros((XUlen, YUlen, ZUlen, 3), dtype=np.float64)
  
      for i in range(len(x1)):
        
        #if x1[i]**2 + y1[i]**2 + z1[i]**2 < 1.0:
        #  bx[i], by[i], bz[i] = 0.0, 0.0, 0.0
        MHDB[ix[i], iy[i], iz[i], :] = [bx[i], by[i], bz[i]]
        MHDposition[ix[i], iy[i], iz[i], :] = [x1[i], y1[i], z1[i]]
  
      # Chunking
      min_chunk = 10
      n_x_split = max(1, XUlen // min_chunk)
      n_y_split = max(1, YUlen // min_chunk)
      n_z_split = max(1, ZUlen // min_chunk)
  
      chunk_x = XUlen // n_x_split
      chunk_y = YUlen // n_y_split
      chunk_z = ZUlen // n_z_split
  
      num_regions = n_x_split * n_y_split * n_z_split
      region_info = []
  
      for r in range(num_regions):
          rx = r % n_x_split
          ry = (r // n_x_split) % n_y_split
          rz = r // (n_x_split * n_y_split)
  
          sx = rx * chunk_x
          sy = ry * chunk_y
          sz = rz * chunk_z
  
          ex = (rx + 1) * chunk_x if rx < n_x_split - 1 else XUlen
          ey = (ry + 1) * chunk_y if ry < n_y_split - 1 else YUlen
          ez = (rz + 1) * chunk_z if rz < n_z_split - 1 else ZUlen
  
          cx = np.mean(XU[sx:ex])
          cy = np.mean(YU[sy:ey])
          cz = np.mean(ZU[sz:ez])
          dist = np.sqrt(cx**2 + cy**2 + cz**2)
  
          region_info.append((dist, r+1, sx+1, ex+1, sy+1, ey+1, sz+1, ez+1))
  
      region_info.sort()
      region_order = [int(r[1]) for r in region_info]
      start_x = [int(r[2]) for r in region_info]
      end_x   = [int(r[3]) for r in region_info]
      start_y = [int(r[4]) for r in region_info]
      end_y   = [int(r[5]) for r in region_info]
      start_z = [int(r[6]) for r in region_info]
      end_z   = [int(r[7]) for r in region_info]
  
      # Convert all arrays to lists
      XU_list = XU.tolist()
      YU_list = YU.tolist()
      ZU_list = ZU.tolist()
  
      MHDposition_list = MHDposition.tolist()
      MHDB_list = MHDB.tolist()
      
      minX_global = np.min(MHDposition[:, :, :, 0])  # Min value in X axis for the whole grid
      maxX_global = np.max(MHDposition[:, :, :, 0])  # Max value in X axis for the whole grid
      minY_global = np.min(MHDposition[:, :, :, 1])  # Min value in Y axis for the whole grid
      maxY_global = np.max(MHDposition[:, :, :, 1])  # Max value in Y axis for the whole grid
      minZ_global = np.min(MHDposition[:, :, :, 2])  # Min value in Z axis for the whole grid
      maxZ_global = np.max(MHDposition[:, :, :, 2])  # Max value in Z axis for the whole grid

      #print("MHD file Loaded")

      del data
      
  
      # Call Fortran
      OTSOLib.mhdstartupsorted(
          XU_list, YU_list, ZU_list,
          MHDposition_list, MHDB_list, n_x_split, n_y_split, n_z_split,
          minX_global,maxX_global,minY_global,maxY_global,minZ_global,maxZ_global,
          region_order, start_x, end_x,
          start_y, end_y, start_z, end_z,
          num_regions,XUlen, YUlen, ZUlen)


import pandas as pd
from datetime import datetime, timedelta
import re
import os

output_directory = os.path.dirname(os.path.realpath(__file__)) 

def extract_solar_wind(current_time):

    current_time = current_time - timedelta(hours=1)
    
    file_path = os.path.join(output_directory, f'space_data.csv')
    df = pd.read_csv(file_path)

    df = df.dropna(how='any')
    
    df['time_tag'] = pd.to_datetime(df['time_tag'])
    df['fetch_timestamp'] = pd.to_datetime(df['fetch_timestamp'])
    
    hourly_avg = hourly_average(df, current_time)

    density_avg = hourly_avg['density'] if 'density' in hourly_avg else None
    speed_avg = hourly_avg['speed'] if 'speed' in hourly_avg else None

    speed_avg = speed_avg*-1

    return speed_avg, density_avg


def extract_magnetic(current_time):

    current_time = current_time - timedelta(hours=1)
    
    file_path = os.path.join(output_directory, f'Magnetic_data.csv')
    df = pd.read_csv(file_path)

    df = df.dropna(how='any')
    
    df['time_tag'] = pd.to_datetime(df['time_tag'])
    df['fetch_timestamp'] = pd.to_datetime(df['fetch_timestamp'])
    

    hourly_avg = hourly_average(df, current_time)

    By_avg = hourly_avg['by_gsm'] if 'by_gsm' in hourly_avg else None
    Bz_avg = hourly_avg['bz_gsm'] if 'bz_gsm' in hourly_avg else None

    return By_avg, Bz_avg


def hourly_average(df, lookup_time):
    lookup_time_ceil = pd.Timestamp(lookup_time).ceil('h')
    
    df_filtered = df[df['time_tag'].dt.floor('h') == lookup_time_ceil]
    
    hourly_avg = df_filtered.mean(numeric_only=True)

    return hourly_avg

def extract_dst_value(file_path, current_time):
    current_day = current_time.day
    current_hour = current_time.hour

    with open(file_path, 'r') as file:
        lines = file.readlines()

    dst_values = {}
    daily_averages = {}

    # Initialize variables to store results in case no data is found
    prev_dst = None
    daily_avg = None

    for line in lines:
        if line.startswith('DST'):
            # Extract year_month and day from the line
            year_month = line[3:7]
            day = int(line[8:10])

            # Check if this line corresponds to the current day
            if day == current_day:
                # Use a regex to extract all numbers, including potential negative and multi-digit values
                dst_data = re.findall(r'-?\d+', line[11:])
                dst_data = dst_data[2:]
                
                # Ignore invalid values (e.g., "999") when extracting hourly data
                hourly_dst_data = [int(val) for val in dst_data[:-1] if int(val) != 999]
                
                # Check if the data for the current hour exists
                if len(hourly_dst_data) > current_hour:
                    prev_dst = hourly_dst_data[current_hour]
                    dst_values[(year_month, day)] = prev_dst

                # Extract the daily average, which is the last value
                try:
                    daily_average = int(dst_data[-1])
                    if daily_average != 999:  # Ignore invalid daily averages
                        daily_averages[(year_month, day)] = daily_average
                except (ValueError, IndexError):
                    pass

    # Retrieve the daily average for the current day if available
    daily_avg = daily_averages.get((year_month, current_day), None)

    return prev_dst, daily_avg

def extract_kp_index(current_time):
    
    kp_index_before = None
    kp_index_after = None
    date_time_before = None
    date_time_after = None

    file_path = os.path.join(output_directory, f'Kp_data.txt')

    with open(file_path, 'r') as file:
        for line in file:
            if line.startswith('#'):
                continue
            
            components = line.split()
            if len(components) < 9:
                continue
            
            date_str = f"{components[0]} {components[1]} {components[2]}"
            time_str = components[3]
            datetime_str = f"{date_str} {time_str}"
            current_datetime = datetime.strptime(datetime_str, '%Y %m %d %H.%M')

            kp_index = float(components[7])

            if current_datetime <= current_time:
                if kp_index_before is None or current_datetime > date_time_before:
                    kp_index_before = kp_index
                    date_time_before = current_datetime
            if current_datetime > current_time:
                if kp_index_after is None or current_datetime < date_time_after:
                    kp_index_after = kp_index
                    date_time_after = current_datetime

    if kp_index_before is not None and kp_index_after is None:
        kp_index_before = round(kp_index_before)
        if kp_index_before >= 6:
            IOPT = 7
        else:
            IOPT = kp_index_before + 1
            IOPT = round(IOPT)

        return kp_index_before, IOPT
    elif kp_index_before is not None and kp_index_after is not None:
        kp_index_before = round(kp_index_before)
        if kp_index_before >= 6:
            IOPT = 7
        else:
            IOPT = kp_index_before + 1
            IOPT = round(IOPT)
        return kp_index_before,IOPT
    else:
        return None



import time
from datetime import datetime
import multiprocessing as mp
import os
from . import fortran_calls, readme_generators,cores, misc, planet_inputs
import pandas as pd
import sys
import queue
import random
import numpy as np
import gc
import csv
from collections import defaultdict

def OTSO_planet(startaltitude,cutoff_comp,minaltitude,maxdistance,maxtime,
           serverdata,livedata,vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,anti,year,
           month,day,hour,minute,second,internalmag,externalmag,
           intmodel,startrigidity,endrigidity,rigiditystep,rigidityscan,
           gyropercent,magnetopause,corenum, azimuth,zenith, asymptotic,asymlevels,unit,
           latstep,longstep,maxlat,minlat,maxlong,minlong,g,h,MHDfile,MHDcoordsys,spheresize,
           array_of_lats_and_longs=None,
           grid_params_user_set=False):

    gc.enable()

    Anum = 1
    PlanetInputArray = planet_inputs.PlanetInputs(startaltitude,cutoff_comp,minaltitude,maxdistance,maxtime,
           serverdata,livedata,vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,Anum,anti,year,
           month,day,hour,minute,second,internalmag,externalmag,
           intmodel,startrigidity,endrigidity,rigiditystep,rigidityscan,
           gyropercent,magnetopause,corenum, azimuth,zenith, asymptotic,asymlevels,unit,
           latstep,longstep,maxlat,minlat,maxlong,minlong,g,h,MHDfile,MHDcoordsys,
           array_of_lats_and_longs=array_of_lats_and_longs,
           grid_params_user_set=grid_params_user_set)

    combined_coordinates = PlanetInputArray[0]
    RigidityArray = PlanetInputArray[1]
    DateArray = PlanetInputArray[2]
    Model = PlanetInputArray[3]
    IntModel = PlanetInputArray[4]
    ParticleArray = PlanetInputArray[5]
    IOPT = PlanetInputArray[6]
    WindArray = PlanetInputArray[7]
    Magnetopause = PlanetInputArray[8]
    MaxStepPercent = PlanetInputArray[9]/100
    EndParams = PlanetInputArray[10]
    Rcomp = PlanetInputArray[11]
    Rscan = PlanetInputArray[12]
    Zenith = PlanetInputArray[13]
    Azimuth = PlanetInputArray[14]
    CoreNum = PlanetInputArray[15]
    asymptotic = PlanetInputArray[16]
    asymlevels = PlanetInputArray[17]
    Alt = PlanetInputArray[18]
    LiveData = PlanetInputArray[19]
    AntiCheck = PlanetInputArray[20]
    g = PlanetInputArray[21]
    h = PlanetInputArray[22]

    del(PlanetInputArray)

    ChildProcesses = []

    totalprocesses = len(combined_coordinates)

    NewCoreNum = misc.CheckCoreNumPlanet(CoreNum)
    FileNamesPlanet = []

    # Generate filenames based on the coordinate pairs
    for coord_pair in combined_coordinates:
        # Ensure formatting is consistent, handle potential floats
        FileNamesPlanet.append(f"{coord_pair[0]:.8f}_{coord_pair[1]:.8f}".replace(".", "dot")) # Using more precision and replacing dots
    
    DataPlanet = []
    i = 1
    # Populate DataPlanet directly from combined_coordinates
    for point, name in zip(combined_coordinates, FileNamesPlanet):
        Core = "Core " + str(i)
        # point[0] is latitude, point[1] is longitude
        DataPlanet.append([name, point[0], point[1], Alt, Zenith, Azimuth, Core]) 
        i = i + 1

    # --- Fix Start: Cap cores and handle empty input --- 
    if totalprocesses == 0:
        print("\nWarning: No coordinate pairs provided or generated. Skipping planet computation.")
        # Need to decide what to return. An empty DataFrame and a basic README?
        EventDate = datetime(year,month,day,hour,minute,second)
        readme = readme_generators.READMEPlanet(None, RigidityArray, EventDate, Model, IntModel, 
                                        AntiCheck, IOPT, WindArray, Magnetopause, 0,
                                        maxlat,maxlong,minlat,minlong, latstep, longstep,
                                        MaxStepPercent*100, EndParams, cutoff_comp, Rscan, 
                                        LiveData, asymptotic, asymlevels, unit, serverdata, kp,
                                        custom_coords_provided=(array_of_lats_and_longs is not None)) # Added flag back temporarily
        return [pd.DataFrame(), readme] # Return empty dataframe and readme
        
    # Ensure the number of processes doesn't exceed the number of points
    actual_cores_to_use = min(NewCoreNum, totalprocesses)
    # --- Fix End --- 

    shuffled_list = DataPlanet.copy()
    random.shuffle(shuffled_list)
    # Use actual_cores_to_use for splitting
    DataLists = np.array_split(shuffled_list, actual_cores_to_use) 

    # Adjust core list generation based on actual_cores_to_use
    CoreList = np.arange(1, actual_cores_to_use + 1) 
    start = time.time()

    current_dir = os.path.dirname(os.path.realpath(__file__))

    planet_list = [get_unique_filename(os.path.join(current_dir, f"Planet{i}.csv")) for i in range(1, CoreNum + 1)]
    for planet_file in planet_list:
        with open(planet_file, mode='a', newline='', encoding='utf-8') as file:  # Open in write ('w') mode
            writer = csv.writer(file)
            
            if asymptotic == "YES":
                asymlevels_with_units = [f"{level} [{unit}]" for level in asymlevels]
                default_headers = ["Latitude", "Longitude", "Rc GV", "Rc Asym"]
                headers = default_headers + asymlevels_with_units
            else:
                headers = ["Latitude", "Longitude", "Rl", "Rc", "Ru"]
            writer.writerow(headers)

    ProcessQueue = mp.Manager().Queue()

    results = []
    resultsfinal = []
    processed = 0
    totalp = 0

    print("OTSO Planet Computation Started")
    sys.stdout.write(f"\r{0:.2f}% complete")
    try:
        if not mp.get_start_method(allow_none=True):
            mp.set_start_method('spawn')
    except RuntimeError:
        pass

    ChildProcesses = []
    for Data, Core, planetfile in zip(DataLists, CoreList, planet_list):
            Child = mp.Process(target=fortran_calls.fortrancallPlanet,  args=(Data, RigidityArray, DateArray, Model, IntModel, 
                                                                              ParticleArray, IOPT, WindArray, 
                                                                              Magnetopause, MaxStepPercent, EndParams, 
                                                                              Rcomp, Rscan, asymptotic, asymlevels, unit,
                                                                              ProcessQueue,g,h,planetfile, MHDfile, MHDcoordsys,
                                                                              spheresize))
            ChildProcesses.append(Child)
        
    for a in ChildProcesses:
        a.start()
 
    while processed < totalprocesses:
        try:
            result_collector = []
            while True:
                try:
                    countint = ProcessQueue.get(timeout=0.001)
                    result_collector.append(countint)
                    processed += 1
                    totalp = totalp + sum(result_collector)
                    result_collector = []
                except queue.Empty:
                    break
    
            
            gc.collect()
            percent_complete = (totalp / totalprocesses) * 100
            sys.stdout.write(f"\r{percent_complete:.2f}% complete ({processed}/{totalprocesses} points)")
            sys.stdout.flush()

    
        except queue.Empty:
            pass
        
        time.sleep(1)

    for b in ChildProcesses:
        b.join()
        b.close()

    processed = 0
    
    ChildProcesses.clear()

    planet = combine_planet_files(planet_list)
   
    print("\nOTSO Planet Computation Complete")
    stop = time.time()
    Printtime = round((stop-start),3)
    print("Whole Program Took: " + str(Printtime) + " seconds")
    
    EventDate = datetime(year,month,day,hour,minute,second)
    
    # --- Fix: Use the first element of DataPlanet for README context --- 
    if not DataPlanet:
         # This case should ideally be caught earlier, but handle defensively
         print("\nError: DataPlanet list is empty before final README generation.")
         readme_context = None 
    else:
        readme_context = DataPlanet[0] # Use the first point's data
    # --- End Fix --- 

    # Pass the representative context to READMEPlanet
    readme = readme_generators.READMEPlanet(readme_context, RigidityArray, EventDate, Model, IntModel, 
                                            AntiCheck, IOPT, WindArray, Magnetopause, Printtime,
                                            maxlat,maxlong,minlat,minlong, latstep, longstep,
                                            MaxStepPercent*100, EndParams, cutoff_comp, Rscan, 
                                            LiveData, asymptotic, asymlevels, unit, serverdata, kp,
                                            custom_coords_provided=(array_of_lats_and_longs is not None))

    if LiveData == 1:
        misc.remove_files()

    return [planet, readme]


def combine_planet_files(planet_list):
    resultsfinal = []
    for x in planet_list:
        df = pd.read_csv(x, header=0)
        for col in df.columns[:3]:
            df[col] = pd.to_numeric(df[col], errors='coerce')
        df = df.dropna(subset=df.columns[:3])
        df = df.drop_duplicates(subset=['Latitude', 'Longitude'])
        resultsfinal.append(df)
        os.remove(x)
    combined_planet = pd.concat(resultsfinal, ignore_index=True)
    combined_planet = combined_planet.drop_duplicates(subset=['Latitude', 'Longitude'])
    combined_planet['Longitude'] = pd.to_numeric(combined_planet['Longitude'], errors='coerce')
    combined_planet['Latitude'] = pd.to_numeric(combined_planet['Latitude'], errors='coerce')
    planet = combined_planet.sort_values(by=["Latitude", "Longitude"], ignore_index=True)
    return planet

def get_unique_filename(filepath):
    """If the file exists, add a numerical suffix to make it unique."""
    base, ext = os.path.splitext(filepath)
    counter = 1
    new_filepath = filepath
    while os.path.exists(new_filepath):
        new_filepath = f"{base}_{counter}{ext}"
        counter += 1
    return new_filepath


from . import MiddleMan as OTSOLib
import time
import os
import shutil
from datetime import datetime
import pandas as pd
import glob
import csv
import numpy as np
import multiprocessing as mp
from . import date
import gc
import sys
import psutil
import tracemalloc
from . import MHDinit

process = psutil.Process()

def fortrancallCutoff(Data, Core, RigidityArray, DateArray, model, IntModel, ParticleArray, IOPT, 
WindArray, Magnetopause, CoordinateSystem, MaxStepPercent, 
EndParams, Rcomp, Rscan, Kp, queue, g, h, MHDfile, MHDCoordSys,spheresize):
    
    if model[1] == 99:
      MHDinit.MHDinitialise(MHDfile)

    for x in Data:
      
      newstart = time.time()
      Position = [x[3],x[1],x[2],x[4],x[5]]


      Station = x[0]

      #print(FileDescriptors)

      NMname = Station

      StartRigidity = RigidityArray[0]
      EndRigidity = RigidityArray[1]
      RigidityStep = RigidityArray[2]
      AtomicNum = ParticleArray[0]
      AntiCheck = ParticleArray[1]

      FileName = NMname + ".csv"
      Rigidities = OTSOLib.cutoff(Position, StartRigidity, EndRigidity, RigidityStep, DateArray, model, IntModel, AtomicNum, AntiCheck, IOPT, WindArray, Magnetopause,
                                   FileName, CoordinateSystem, MaxStepPercent, EndParams, Rcomp, Rscan, g, h, MHDCoordSys,spheresize)

      
      Rigiditydataframe = pd.DataFrame({Station: Rigidities}, index=['Ru', 'Rc', 'Rl'])
      queue.put(Rigiditydataframe)

    return

def fortrancallCone(Data, Core, RigidityArray, DateArray, model, IntModel, ParticleArray, IOPT, WindArray, Magnetopause,
                     CoordinateSystem, MaxStepPercent, EndParams, queue, g, h, MHDfile, MHDCoordSys,spheresize):
    
    if model[1] == 99:
      MHDinit.MHDinitialise(MHDfile)
    
    for x in Data:
      
      Position = [x[3],x[1],x[2],x[4],x[5]]
      Station = x[0]

      NMname = Station

      StartRigidity = RigidityArray[0]
      EndRigidity = RigidityArray[1]
      RigidityStep = RigidityArray[2]
      AtomicNum = ParticleArray[0]
      AntiCheck = ParticleArray[1]

      length = int(StartRigidity/RigidityStep)

      FileName = NMname + ".csv"
      Cone, Rigidities = OTSOLib.cone(Position, StartRigidity, EndRigidity, RigidityStep, DateArray, model, IntModel, AtomicNum, AntiCheck, IOPT, WindArray,
                                       Magnetopause, FileName, CoordinateSystem, MaxStepPercent, EndParams, length, g, h, MHDCoordSys,spheresize)
      decoded_lines = [line.decode('utf-8').strip() for sublist in Cone for line in sublist]

      data = [line.split() for line in decoded_lines]

      Conedf = pd.DataFrame(data, columns=['R [GV]', 'Filter', 'ALat', 'ALong'])
      Conedf[NMname] = Conedf[['Filter', 'ALat', 'ALong']].astype(str).agg(';'.join, axis=1)
      Conedf.drop(columns=['Filter', 'ALat', 'ALong'], inplace=True)

      Rigiditydataframe = pd.DataFrame({Station: Rigidities}, index=['Ru', 'Rc', 'Rl'])
 
      queue.put([Conedf, Rigiditydataframe])
    
    return


def fortrancallPlanet(Data, Rigidity, DateArray, model, IntModel, ParticleArray, IOPT, WindArray, Magnetopause, MaxStepPercent, 
                      EndParams, Rcomp, Rscan, asymptotic, asymlevels, unit, queue, g, h, PlanetFile,MHDfile, MHDCoordSys,spheresize):
  with open(PlanetFile, mode='a', newline='', encoding='utf-8') as file:
    writer = csv.writer(file)

    if model[1] == 99:
      MHDinit.MHDinitialise(MHDfile)

    for x in Data:
        
        Position = [x[3],x[1],x[2],x[4],x[5]]
        Station = x[0]

        AtomicNum = ParticleArray[0]
        AntiCheck = ParticleArray[1]
  
        NMname = Station
        Rigidities = [0,0,0]
  
        FileName = NMname + ".csv"
        Rigidities = OTSOLib.planet(Position, Rigidity, DateArray, model, IntModel, AtomicNum, AntiCheck, IOPT, WindArray, Magnetopause, FileName,
                                     MaxStepPercent, EndParams, Rcomp, Rscan, g, h, MHDCoordSys,spheresize)
        CoordinateSystem = "GEO"
  
        lat_long_pairs = []
        P_List = []
  
        if asymptotic == "YES":
          Energy_List = asymlevels.copy()
          if unit == "GeV":   
            E_0 = 0.938
            for i in Energy_List:
              R = (i**2 + 2*i*E_0)**(0.5)
              P_List.append(R)
            P_List.insert(0, Rigidities[1])
          else:
            if unit == "GV": 
              P_List = Energy_List.copy()
            P_List.insert(0, Rigidities[1])
          for P in P_List:
              if P == P_List[0]:  # Check if it's the first value in P_List
                  while True:
                      bool, Lat, Long = OTSOLib.trajectory(Position, P, DateArray, model, IntModel, AtomicNum, AntiCheck, IOPT, WindArray, Magnetopause, FileName,
                                                            CoordinateSystem, MaxStepPercent, EndParams, g, h, MHDCoordSys,spheresize)
                      P += Rigidity[2]
                      if bool == 1:
                          lat_long_pairs.append([bool, round(Lat, 3), round(Long, 3)])
                          break 
              else:
                  bool, Lat, Long = OTSOLib.trajectory(Position, P, DateArray, model, IntModel, AtomicNum, AntiCheck, IOPT, WindArray, Magnetopause, FileName,
                                                        CoordinateSystem, MaxStepPercent, EndParams, g, h, MHDCoordSys,spheresize)
                  lat_long_pairs.append([bool, round(Lat, 3), round(Long, 3)])
  
          formatted_list = [f"{bool};{lat};{long}" for bool, lat, long in lat_long_pairs]
  
          formatted_list.insert(0, Position[1])
          formatted_list.insert(1, Position[2])
          formatted_list.insert(2, Rigidities[1])
          writer.writerow(formatted_list)
          queue.put(1)
  
        else:
           data = [x[1],x[2],Rigidities[0],Rigidities[1], Rigidities[2]]
           writer.writerow(data)
           queue.put(1)
           
    file.close()

  return

  

def fortrancallTrajectory(Data, Core, Rigidity, DateArray, model, IntModel, ParticleArray, IOPT, WindArray, 
                          Magnetopause, CoordinateSystem, MaxStepPercent, EndParams, queue, g, h, MHDfile, MHDCoordSys,spheresize):
  
  if model[1] == 99:
    MHDinit.MHDinitialise(MHDfile)
  
  for x in Data:
    
    Position = [x[3],x[1],x[2],x[4],x[5]]
    Station = x[0]

    AtomicNum = ParticleArray[0]
    AntiCheck = ParticleArray[1]

    NMname = Station

    FileName = NMname + ".csv"
    OTSOLib.trajectory_full(Position, Rigidity, DateArray, model, IntModel, AtomicNum, AntiCheck, IOPT, WindArray, Magnetopause, FileName, 
                            CoordinateSystem, MaxStepPercent, EndParams, g, h, MHDCoordSys,spheresize)
    Trajectory = pd.read_csv(FileName)
    Trajectory.columns = [f"{col}_Re [{CoordinateSystem}]" for col in Trajectory.columns]
    
    dataframe_dict = {NMname: Trajectory}

    queue.put(dataframe_dict)
    os.remove(FileName)

  return

def fortrancallFlight(Data, Rigidity, DateArray, model, IntModel, ParticleArray, IOPT, WindArray, Magnetopause, 
                      MaxStepPercent, EndParams, Rcomp, Rscan, asymptotic, asymlevels, unit, queue, g, h, 
                      CoordinateSystem, FlightFile, MHDfile, MHDCoordSys,spheresize):
  with open(FlightFile, mode='a', newline='', encoding='utf-8') as file:
    if asymptotic == "YES":
        asymlevels_with_units = [f"{level} [{unit}]" for level in asymlevels]
        defualt_headers = ["Date","Latitude","Longitude","Altitude","Rc GV","Rc Asym"]
        headers = defualt_headers + asymlevels_with_units
    else:
        headers = ["Date","Latitude", "Longitude","Altitude", "Ru", "Rc", "Rl"]

    writer = csv.writer(file)
    writer.writerow(headers)

    if model[1] == 99:
      MHDinit.MHDinitialise(MHDfile)


    for x,y,z,I in zip(Data,DateArray,WindArray,IOPT):
        
        Position = [x[3],x[1],x[2],x[4],x[5]]
        Station = x[0]
        
        datetimeobj = date.convert_to_datetime(y)
        Wind = z
  
        StartRigidity = Rigidity[0]
        EndRigidity = Rigidity[1]
        RigidityStep = Rigidity[2]
        AtomicNum = ParticleArray[0]
        AntiCheck = ParticleArray[1]
        AtomicNum = ParticleArray[0]
        AntiCheck = ParticleArray[1]
  
        NMname = Station
        Rigidities = [0,0,0]
  
        FileName = NMname + ".csv"
        Rigidities = Rigidities = OTSOLib.cutoff(Position, StartRigidity, EndRigidity, RigidityStep, y, model, IntModel, AtomicNum, AntiCheck, I, Wind, Magnetopause, 
                                                 FileName, CoordinateSystem, MaxStepPercent, EndParams, Rcomp, Rscan, g, h, MHDCoordSys,spheresize)
  
        lat_long_pairs = []
        P_List = []
  
        if asymptotic == "YES":
          Energy_List = asymlevels.copy()
          if unit == "GeV":   
            E_0 = 0.938
            for i in Energy_List:
              R = (i**2 + 2*i*E_0)**(0.5)
              P_List.append(R)
            P_List.insert(0, round(Rigidities[1], 5))
          else:
            if unit == "GV": 
              P_List = Energy_List.copy()
            P_List.insert(0, Rigidities[1])
          for P in P_List:
              if P == P_List[0]:
                  while True:
                      bool, Lat, Long = OTSOLib.trajectory(Position, P, y, model, IntModel, AtomicNum, AntiCheck, I, Wind, Magnetopause, FileName, 
                                                           CoordinateSystem, MaxStepPercent, EndParams, g, h, MHDCoordSys,spheresize)
                      P += RigidityStep
                      if bool == 1:
                          lat_long_pairs.append([bool, round(Lat, 3), round(Long, 3)])
                          break 
              else:
                  bool, Lat, Long = OTSOLib.trajectory(Position, P, y, model, IntModel, AtomicNum, AntiCheck, I, Wind, Magnetopause, FileName, 
                                                       CoordinateSystem, MaxStepPercent, EndParams, g, h, MHDCoordSys,spheresize)
                  lat_long_pairs.append([bool, round(Lat, 3), round(Long, 3)])
          
          asymlevels_with_units = [f"{level} [{unit}]" for level in asymlevels]
          defualt_headers = ["Date","Latitude","Longitude","Altitude","Rc GV","Rc Asym"]
          headers = defualt_headers + asymlevels_with_units
          df = pd.DataFrame(columns=headers)
  
          formatted_list = [f"{bool};{lat};{long}" for bool, lat, long in lat_long_pairs]
  
          formatted_list.insert(0, datetimeobj)
          formatted_list.insert(1, round(float(Position[1]),5))
          formatted_list.insert(2, round(float(Position[2]),5))
          formatted_list.insert(3, round(float(Position[0]),5))
          formatted_list.insert(4, round(Rigidities[1],5))
          writer.writerow(formatted_list)
          queue.put(1)
  
        else:
           headers = ["Date","Latitude", "Longitude","Altitude", "Ru", "Rc", "Rl"]
           data = [datetimeobj,x[1],x[2],x[3],Rigidities[0],Rigidities[1], Rigidities[2]]
           writer.writerow(data)
           queue.put(1)
    file.close()
  return


def fortrancallMagfield(Data, DateArray, model, IOPT, WindArray, CoordinateSystem, queue,g,h,MHDfile, MHDCoordSys):
    
  if model[1] == 99:
    MHDinit.MHDinitialise(MHDfile)

  for x in Data:
    Position = x
    
    Bfield = OTSOLib.magstrength(Position, DateArray, model, IOPT, WindArray, CoordinateSystem,MHDCoordSys, g, h)
    Bfield = Bfield*10**9
    combined_array = np.concatenate((Position, Bfield))
    coord_suffix = f"_{CoordinateSystem}"  # Add the coordinate system suffix
    columns = [f"X{coord_suffix} [Re]", f"Y{coord_suffix} [Re]", f"Z{coord_suffix} [Re]", f"GSM_Bx [nT]", f"GSM_By [nT]", f"GSM_Bz [nT]"]
    df = pd.DataFrame(columns=columns)
    df.loc[len(df)] = combined_array
    queue.put(df)

  return

def fortrancallCoordtrans(Data, DateArray, CoordIN, CoordOUT, queue):
    for x,y in zip(Data,DateArray):
      Position = [1+(x[2]/6371.0),x[0],x[1]]

      datetimeobj = date.convert_to_datetime(y)

      year = y[0]
      day = y[1]
      hour = y[2]
      minute = y[3]
      secint = y[4]
      sectot = y[5]
      
      Coords = OTSOLib.coordtrans(Position,year,day,hour,minute,secint,sectot,CoordIN,CoordOUT)
      combined_array = np.concatenate(([datetimeobj], Position, Coords))

      coord_suffix = f"_{CoordIN}"
      coord_suffix2 = f"_{CoordOUT}"
      columns = ["Date",f"X{coord_suffix} [Re]", f"Y{coord_suffix} [Re]", f"Z{coord_suffix} [Re]", f"X{coord_suffix2} [Re]", f"Y{coord_suffix2} [Re]", f"Z{coord_suffix2} [Re]"]
      if CoordOUT == "GDZ" or CoordOUT == "SPH":
         columns = ["Date",f"X{coord_suffix} [Re]", f"Y{coord_suffix} [Re]", f"Z{coord_suffix} [Re]", f"altitude{coord_suffix2} [km]", f"latitude{coord_suffix2}", f"longitude{coord_suffix2}"]
      df = pd.DataFrame(columns=columns)
      df.loc[len(df)] = combined_array
      queue.put(df)

    return

def fortrancallTrace(Data, Rigidity, DateArray, model, IntModel, ParticleArray, IOPT, WindArray, Magnetopause, 
                     CoordinateSystem, MaxStepPercent, EndParams, queue, g, h, MHDfile, MHDCoordSys,spheresize):
    
    if model[1] == 99:
      MHDinit.MHDinitialise(MHDfile)
    
    for x in Data:
      
      Position = [x[3],x[1],x[2],x[4],x[5]]

      AtomicNum = ParticleArray[0]
      AntiCheck = ParticleArray[1]

      Filename = f"{x[1]}_{x[2]}.csv"
      name = f"{x[1]}_{x[2]}"

      OTSOLib.fieldtrace(Position, Rigidity, DateArray, model, IntModel, AtomicNum, AntiCheck, IOPT, WindArray, Magnetopause, CoordinateSystem, MaxStepPercent,
                          EndParams, Filename, g, h, MHDCoordSys,spheresize)
      Trace = pd.read_csv(Filename)
      coordsystem2 = "GSM"
      columns = Trace.columns

      new_columns = [f"{col}_{CoordinateSystem} [Re]" if i < 3 else f"{col}_{coordsystem2} [nT]" for i, col in enumerate(columns)]
      if CoordinateSystem == "GDZ" or CoordinateSystem == "SPH":
        column_names = ["alt [km]", "latitude", "longitude"]
        new_columns = [f"{col}_{column_names[i]}" if i < 3 else f"{col}_{coordsystem2} [nT]" for i, col in enumerate(columns)]

      Trace.columns = new_columns

      Trace.columns = new_columns
            
      dataframe_dict = {name: Trace}
  
      queue.put(dataframe_dict)
      os.remove(Filename)
    
    return


import ftplib
import pandas as pd
import csv, math
from datetime import datetime, timedelta
import shutil
import os
from ftplib import FTP_TLS
import numpy as np

def download_omni_data(year):
    ftps = FTP_TLS('spdf.gsfc.nasa.gov')
    ftps.login()
    ftps.prot_p()
    ftps.cwd('/pub/data/omni/high_res_omni/')
    files = ftps.nlst()
    file_name = f'omni_5min{year}.asc'
    
    if file_name not in files:
        print(f"File for year {year} not found.")
        ftps.quit()
        return

    # Get the directory of the current script
    script_dir = os.path.join(os.path.dirname(__file__), "")
    save_path = os.path.join(script_dir, f'omni_5min_{year}.lst')

    with open(save_path, 'wb') as f:
        ftps.retrbinary(f'RETR {file_name}', f.write)

    ftps.quit()

    parse_and_convert_to_csv_high_res(save_path, f'omni_{year}_high_res.csv')


def download_omni_low_res_data(year):
    ftps = FTP_TLS('spdf.gsfc.nasa.gov')
    ftps.login()
    ftps.prot_p()

    ftps.cwd('/pub/data/omni/low_res_omni/')

    files = ftps.nlst()

    file_name = f'omni2_{year}.dat'
    file_name2 = f'omni2_{year+1}.dat'

    endfile_name = f'functions/omni2_{year}.dat'
    endfile_name2 = f'functions/omni2_{year+1}.dat'

    csvfile_name = f'omni_{year}_low_res.csv'
    csvfile_name2 = f'omni_{year+1}_low_res.csv'

    script_dir = os.path.dirname(os.path.dirname(__file__))
    file_path = os.path.join(script_dir, endfile_name)
    file_path2 = os.path.join(script_dir, endfile_name2)

    csvfile_path = os.path.join(script_dir, csvfile_name)
    csvfile_path2 = os.path.join(script_dir, csvfile_name2)
    
    if file_name not in files:
        print(f"File for year {year} not found.")
        ftps.quit()
        return
    
    if file_name2 in files:
        with open(file_path2, 'wb') as f:
            ftps.retrbinary(f'RETR {file_name2}', f.write)
        parse_and_convert_to_csv_low_res(file_path2, f'omni_{year+1}_low_res.csv')

    with open(file_path, 'wb') as f:
        ftps.retrbinary(f'RETR {file_name}', f.write)

    ftps.quit()

    parse_and_convert_to_csv_low_res(file_path, f'omni_{year}_low_res.csv')



def convert_to_datetime(year, decimal_day, hour, minute):
    """ Convert year, decimal day, and hour into a datetime string. """
    year = int(year)
    day_of_year = int(float(decimal_day))
    hour = int(hour)
    minute = int(minute)


    initial_date = datetime(year, 1, 1) + timedelta(days=day_of_year - 1)
    datetime_obj = initial_date + timedelta(hours=hour,minutes=minute)

    return datetime_obj.strftime('%Y-%m-%d %H:%M:%S')

def parse_and_convert_to_csv_high_res(input_file, output_file):
    """ Parse high-res OMNI data file and convert to CSV with specific columns. """
    with open(input_file, 'r') as datfile:
        lines = datfile.readlines()

    FILLER_VALUES = {"99999", "9999999", "-999.9", "9999.99", "9999.9", "999.9", "99999.0", "-1.00e+05", "999999", "999.99"}


    rows = [line.split() for line in lines]

    output_headers = ["Date", "By", "Bz", "V", "Density", "Pdyn"]

    # Get the directory of the current script
    script_dir = os.path.dirname(os.path.realpath(__file__))
    output_path = os.path.join(script_dir, output_file)

    with open(output_path, mode='w', newline='') as csvfile:
        csv_writer = csv.writer(csvfile)
        csv_writer.writerow(output_headers)

        for row in rows:
            try:
                datetime_str = convert_to_datetime(row[0], row[1], row[2],row[3])  # Year, DOY, Hour, Minute
                By_value = row[17]
                Bz_value = row[18]
                V_value = row[21]
                Density_value = row[25]
                Pdyn_value = row[27]

                values_to_check = {By_value, Bz_value, V_value, Density_value, Pdyn_value}

                # Skip the row if any value is a filler
                if any(val in FILLER_VALUES for val in values_to_check):
                    continue

                csv_writer.writerow([datetime_str, By_value, Bz_value, V_value, Density_value, Pdyn_value])
            except (IndexError, ValueError):
                continue

def parse_and_convert_to_csv_low_res(input_file, output_file):
    """ Parse the data file and convert it to a CSV with only specific columns. """
    with open(input_file, 'r') as datfile:
        lines = datfile.readlines()
        
        rows = [line.split() for line in lines]
        
    output_headers = ["Date", "Kp", "Dst", "By", "Bz", "V", "Density", "Pdyn"]

    # Get the directory of the current script
    script_dir = os.path.dirname(os.path.realpath(__file__))
    output_path = os.path.join(script_dir, output_file)
    
    with open(output_path, mode='w', newline='') as csvfile:
        csv_writer = csv.writer(csvfile)
        csv_writer.writerow(output_headers)

        for row in rows:
            try:
                datetime_str = convert_to_datetime(row[0], row[1], row[2],0)
                kp_value = process_kp_value(row[38])
                dst_value = row[40]
                By_value = row[15]
                Bz_value = row[16]
                V_value = row[24]
                Density_value = row[23]
                Pdyn_value = row[28]

                csv_writer.writerow([datetime_str, kp_value, dst_value, By_value, Bz_value, V_value, Density_value, Pdyn_value])
            
            except (IndexError, ValueError):
                continue

def parse_and_convert_to_csv(input_file, output_file):
    """ Parse the data file and convert it to a CSV with only specific columns. """
    with open(input_file, 'r') as datfile:
        lines = datfile.readlines()
        
        rows = [line.split() for line in lines]
        
    output_headers = ["Date", "Kp", "Dst", "By", "Bz", "V", "Density", "Pdyn"]

    # Get the directory of the current script
    script_dir = os.path.dirname(os.path.realpath(__file__))
    output_path = os.path.join(script_dir, output_file)
    
    with open(output_path, mode='w', newline='') as csvfile:
        csv_writer = csv.writer(csvfile)
        csv_writer.writerow(output_headers)

        for row in rows:
            try:
                datetime_str = convert_to_datetime(row[0], row[1], row[2],0)
                kp_value = process_kp_value(row[38])
                dst_value = row[40]
                By_value = row[15]
                Bz_value = row[16]
                V_value = row[24]
                Density_value = row[23]
                Pdyn_value = row[28]

                csv_writer.writerow([datetime_str, kp_value, dst_value, By_value, Bz_value, V_value, Density_value, Pdyn_value])
            
            except (IndexError, ValueError):
                continue

def process_kp_value(kp_value):
    try:
        kp_value = int(kp_value)
        rounded_kp = round(kp_value, -1) 
        return int(rounded_kp / 10) 
    except ValueError:
        return kp_value

def extract_row_from_csv(output_file_path, target_datetime):
    target_datetime_str = target_datetime.strftime('%d/%m/%y %H:00:00')

    with open(output_file_path, 'r') as csvfile:
        csv_reader = csv.DictReader(csvfile)

        for row in csv_reader:
            if row['Datetime'] == target_datetime_str:
                validate_row(row)
                return row

    print(f"No matching row found for the given datetime: {target_datetime_str}")
    return None

def validate_row(row):
    for key in row:
        if row[key] == 'NaN' or row[key] == '9999.' or row[key] == '999.9':
            print("ERROR: No valid OMNI data found for given date. Please check other sources. \nOTSO will now terminate.")
            exit()

def PullOMNI(year):
    download_omni_data(year)
    download_omni_low_res_data(year)

def PullOMNILowRes(year):
    download_omni_low_res_data(year)

def OMNI_to_csv(year):
    headers = [
        'BXGSM', 'By', 'Bz', 'V', 
        'VYGSE', 'VZGSE', 'Density', 'TEMP', 'IMFFLAG', 'ISWFLAG', 'TILT', 
        'Pdyn', 'W1', 'W2', 'W3', 'W4', 'W5', 'W6', 'Date'
    ]

    file = os.path.join(os.path.dirname(__file__), f'{year}_OMNI_5m_with_TS05_variables.dat')
    data = pd.read_csv(file, sep=r'\s+', header=None)

    IYEAR = data[0]
    IDAY = data[1]
    IHOUR = data[2]
    MIN = data[3]

    def split_column(value):
       if '*' in str(value):
           return value.split('*')[0], 0.0
       return float(value), None

    split_values = data[13].apply(split_column)
    data[13] = split_values.apply(lambda x: x[0])

    # Check if any value contains an asterisk and add new column if necessary
    if split_values.apply(lambda x: x[1]).notna().any():
        data.insert(14, 'new_column', split_values.apply(lambda x: x[1]).fillna(0.0))

    # Convert the original column back to float
    data[13] = data[13].astype(float)
    
    data['Date'] = [
        datetime(int(year), 1, 1) + pd.to_timedelta(int(day)-1 if int(day) > 0 else 0, unit='d') + \
        pd.to_timedelta(int(hour), unit='h') + pd.to_timedelta(int(minute), unit='m')
        for year, day, hour, minute in zip(IYEAR, IDAY, IHOUR, MIN)
    ]

    data = data.drop(columns=[0, 1, 2, 3])
    data.columns = headers

    columns_to_remove = ['BXGSM', 'VYGSE', 'VZGSE', 'TEMP', 'IMFFLAG', 'ISWFLAG', 'TILT']
    data = data.drop(columns=columns_to_remove, errors='ignore')

    remaining_columns = [col for col in headers if col != 'Date' and col not in columns_to_remove]
    data = data[['Date'] + remaining_columns] 
    
    csv_filename = os.path.join(os.path.dirname(__file__), f'{year}_TSY_Data.csv')
    data.to_csv(csv_filename, index=False)


def TSY01(File):
    filepath = os.path.join(os.path.dirname(__file__), File)
    data = pd.read_csv(filepath)

    data['Date'] = pd.to_datetime(data['Date'])

    results = []
    G1_values = []
    G2_values = []
    G3_values = []

    for i in range(len(data)):
        current_time = data['Date'].iloc[i]
        start_index = max(0, i - 11)
        relevant_rows = data.iloc[start_index:i + 1]
        relevant_rows = relevant_rows[relevant_rows['Date'] > (current_time - pd.Timedelta(hours=1))]
    

        G1,G2,G3 = TSY01_Constants(relevant_rows)
        G1_values.append(G1)
        G2_values.append(G2)
        G3_values.append(G3)

    pdyn_index = data.columns.get_loc('Pdyn') + 1

    data.insert(pdyn_index, 'G1', G1_values)
    data.insert(pdyn_index + 1, 'G2', G2_values)
    data.insert(pdyn_index + 2, 'G3', G3_values)

    data.to_csv(filepath, index=False)
    

def TSY01_Constants(data):

    IMFy = data["By"]
    IMFz = data["Bz"]
    Speed = data["V"]
    Density = data["Density"]
    G1 = 0
    G2 = 0
    G3 = 0

    for (By, Bz, V, N) in zip(IMFy, IMFz, Speed, Density):
        By = By
        Bz = Bz
        if V < 0:
            V = -1*V
        W = 1/len(IMFy)
    
    
        B = (By*By + Bz*Bz)**(0.5)
        h = (((B/40)**(2))/(1 + B/40))
    
        if(By == 0 and Bz == 0):
            phi = 0
        else:
            phi = math.atan2(By,Bz)
            if(phi <= 0):
                phi = phi + 2*math.pi
    
        if(Bz < 0):
            Bs = abs(Bz)
        elif Bz >= 0:
            Bs = 0
        
        G1 = G1 + (W)*V*h*(math.sin(phi/2)**3)
        G2 = G2 + (0.005)*(W)*(V)*Bs
        G3 = G3 + (N*V*Bs)/2000

    return round(G1, 2), round(G2, 2), round(G3, 2)

def CombineLowRes(input_file, year):
    # Read the CSV file
    df = pd.read_csv(input_file)
    
    # Define new headers
    new_headers = ['Date', 'By', 'Bz', 'V', 'Density', 'Pdyn', 'Dst', 'Kp', 
                   'G1', 'G2', 'G3', 'W1', 'W2', 'W3', 'W4', 'W5', 'W6']
    
    # Create a new DataFrame with the required structure
    new_df = pd.DataFrame(columns=new_headers)
    
    # Copy Date and Kp values
    new_df['Date'] = df['Date']
    new_df['Kp'] = df['Kp']
    
    # Fill all other columns with zero
    for col in new_headers:
        if col not in ['Date', 'Kp']:
            new_df[col] = 0

    new_df.to_csv(f'{year}_TSY_Inputs.csv', index=False)
    source_file = f'{year}_TSY_Inputs.csv'
    destination_folder = os.path.join(os.path.dirname(__file__), "ServerData")
    os.makedirs(destination_folder, exist_ok=True)
    destination_file = os.path.join(destination_folder, os.path.basename(source_file))
    shutil.move(source_file, destination_file)
    
import os
import pandas as pd
import shutil

def Combine(TSYfile, high_res_file, low_res_file, year):
    base_dir = os.path.dirname(__file__)
    TSYfile = os.path.join(base_dir, TSYfile)
    high_res_file = os.path.join(base_dir, high_res_file)
    low_res_file = os.path.join(base_dir, low_res_file)
    futurefile = os.path.join(base_dir, f'omni_{year+1}_low_res.csv')

    # Load all data
    partial_df = pd.read_csv(TSYfile, parse_dates=['Date']).set_index('Date')
    high_res_df = pd.read_csv(high_res_file, parse_dates=['Date']).set_index('Date')
    low_res_df = pd.read_csv(low_res_file, parse_dates=['Date']).set_index('Date')
    low_res_5min = low_res_df.resample('5min').ffill()

    
    if 'Dst' in low_res_df.columns:
        low_res_5min['Dst'] = (low_res_df['Dst'].resample('5min').interpolate(method='linear').round().astype(int)
        )

    # Define full 5-minute time range for the year
    start_time = partial_df.index.min()
    end_time = partial_df.index.max()
    full_index = pd.date_range(start=start_time, end=end_time, freq='5min')

    # Reindex and combine with priority: partial > high_res > low_res
    combined_df = pd.DataFrame(index=full_index)
    combined_df = combined_df.combine_first(partial_df)
    combined_df = combined_df.combine_first(high_res_df)
    combined_df = combined_df.combine_first(low_res_5min)

    # Reset index and fix column order
    combined_df = combined_df.reset_index().rename(columns={'index': 'Date'})

    # Ensure V is positive
    if 'V' in combined_df.columns:
        combined_df['V'] = combined_df['V'].abs()

    # Fill final Dec 31 values from following year if available
    if os.path.exists(futurefile):
        future_df = pd.read_csv(futurefile)
        dst_value = future_df.loc[0, 'Dst'] if 'Dst' in future_df.columns else None
        kp_value = future_df.loc[0, 'Kp'] if 'Kp' in future_df.columns else None
        if dst_value is not None and kp_value is not None:
            mask = (combined_df['Date'].dt.year == year) & \
                   (combined_df['Date'].dt.month == 12) & \
                   (combined_df['Date'].dt.day == 31) & \
                   (combined_df['Date'].dt.hour == 23) & \
                   (combined_df['Date'].dt.minute >= 5)
            combined_df.loc[mask, 'Dst'] = dst_value
            combined_df.loc[mask, 'Kp'] = kp_value

    # Ensure column completeness and order
    desired_order = ['Date', 'By', 'Bz', 'V', 'Density', 'Pdyn', 'Dst', 'Kp',
                     'G1', 'G2', 'G3', 'W1', 'W2', 'W3', 'W4', 'W5', 'W6']
    for col in desired_order:
        if col not in combined_df.columns:
            combined_df[col] = pd.NA
    combined_df = combined_df[desired_order]
    combined_df = combined_df.drop_duplicates(subset='Date', keep='first')

    # Write output CSV
    output_file = os.path.join(base_dir, f'{year}_TSY_Inputs.csv')
    combined_df.to_csv(output_file, index=False)

    # Move to ServerData folder
    destination_folder = os.path.join(base_dir, "ServerData")
    os.makedirs(destination_folder, exist_ok=True)
    shutil.move(output_file, os.path.join(destination_folder, os.path.basename(output_file)))




def Omnidelete(OMNIYEAR):

    current_directory = os.path.join(os.path.dirname(__file__), "")

    file1 = f'{OMNIYEAR}_TSY_Data.csv'
    file2 = f'omni_{OMNIYEAR}_low_res.csv'
    file3 = f'omni_5min_{OMNIYEAR}.lst'
    file4 = f'omni2_{OMNIYEAR}.dat'
    file5 = f'{OMNIYEAR}_IMF_&_SW_gaps_le_3hrs_filled.txt'
    file6 = f'{OMNIYEAR}_IMF_gaps_le_3hrs_filled.txt'
    file7 = f'{OMNIYEAR}_Interval_list.txt'
    file8 = f'{OMNIYEAR}_OMNI_5m_with_TS05_variables.dat'
    file9 = f'omni_{OMNIYEAR+1}_low_res.csv'
    file10 = f'omni2_{OMNIYEAR+1}.dat'
    file11 = f'omni_{OMNIYEAR}_high_res.csv'
    
    if OMNIYEAR < datetime.now().year:
        filelist = [file1,file2,file3,file4,file5,file6,file7,file8,file9,file10,file11]
    else:
        filelist = [file1,file2,file3,file4,file5,file6,file7,file8]

    for i in filelist:
        try:
            os.remove(os.path.join(current_directory, i))
        except Exception as e:
            print(f"Error: {e}")

    if os.path.exists(file9):
        os.remove(os.path.join(current_directory, file9))

def OmnideleteLowRes(OMNIYEAR):

    current_directory = os.path.dirname(__file__)

    file2 = f'omni_{OMNIYEAR}_low_res.csv'
    file3 = f'omni_{OMNIYEAR+1}_low_res.csv'
    file4 = f'omni2_{OMNIYEAR+1}.dat'
    file5 = f'omni2_{OMNIYEAR}.dat'
    
    if OMNIYEAR < datetime.now().year:
        filelist = [file2,file3,file4,file5]
    else:
        filelist = [file2,file3,file5]

    for i in filelist:
        try:
            os.remove(os.path.join(current_directory, i))
        except Exception as e:
            print(f"Error: {e}")

    if os.path.exists(file3):
        os.remove(os.path.join(current_directory, file3))



def coordtrans(Locations,dates,CoordIN,CoordOUT,corenum=None):
    from .Parameters.functions import otso_coordtrans
    import psutil

    if corenum is None:
       corenum = psutil.cpu_count(logical=False) - 2
       if corenum <= 0:
          corenum = 1
    
    arguments = locals()
    for arg in arguments:
       if arguments[arg] is None:
          arguments[arg] = []

    coordtrans = otso_coordtrans.OTSO_coordtrans(Locations,dates,CoordIN,CoordOUT,corenum)
    
    return coordtrans


def flight(latitudes, longitudes, dates,altitudes,cutoff_comp="Vertical",minaltitude=20,maxdistance=100,maxtime=0,
           serverdata="OFF",livedata="OFF",vx=None,vy=None,vz=None,by=None,bz=None,density=None,pdyn=None,Dst=None,
           G1=None,G2=None,G3=None,W1=None,W2=None,W3=None,W4=None,W5=None,W6=None,kp=None,Anum=1,anti="YES",internalmag="IGRF",externalmag="TSY89",
           intmodel="Boris",startrigidity=20,endrigidity=0,rigiditystep=0.01,rigidityscan="ON",
           coordsystem="GEO",gyropercent=15,magnetopause="Kobel",corenum=None,azimuth=0,zenith=0,g=None,h=None,
           asymptotic="NO",asymlevels=[0.1,0.3,0.5,1,2,3,4,5,6,7,8,9,10,15,20,30,50,70,100,300,500,700,1000],
           unit="GeV",MHDfile=None,MHDcoordsys=None,spheresize=25):
    from .Parameters.functions import otso_flight
    import psutil

    if corenum is None:
       corenum = psutil.cpu_count(logical=False) - 2
       if corenum <= 0:
          corenum = 1
       
    arguments = locals()

    default_values = {
        "vx": -500, "vy": 0, "vz": 0, "by": 5.0, "bz": 5.0, "density": 1, "pdyn": 0, "Dst": 0,
        "G1": 0.00, "G2": 0.00, "G3": 0.00, "W1": 0, "W2": 0, "W3": 0, "W4": 0, "W5": 0, "W6": 0, "kp": 0
    }

    for arg in arguments:
         if arguments[arg] is None:
             arguments[arg] = []


    if serverdata != "ON" and livedata != "ON":
        for var_name, default_value in default_values.items():
            if arguments[var_name] is None or not arguments[var_name]:  # If None or empty list
                arguments[var_name] = [default_value] * len(latitudes)


    
    flight = otso_flight.OTSO_flight(latitudes, longitudes, dates, altitudes, cutoff_comp, minaltitude, maxdistance, maxtime,
             serverdata, livedata, arguments["vx"], arguments["vy"], arguments["vz"], arguments["by"], arguments["bz"], 
             arguments["density"], arguments["pdyn"], arguments["Dst"], arguments["G1"], arguments["G2"], arguments["G3"], 
             arguments["W1"], arguments["W2"], arguments["W3"], arguments["W4"], arguments["W5"], arguments["W6"], 
             arguments["kp"], Anum, anti, internalmag, externalmag, intmodel, startrigidity, endrigidity, rigiditystep, 
             rigidityscan, coordsystem, gyropercent, magnetopause, corenum, azimuth, zenith, arguments["g"], arguments["h"], 
             asymptotic, asymlevels, unit,MHDfile,MHDcoordsys,spheresize
             )

    
    return flight


import sys, os, csv
import subprocess
import platform
import shutil
import pandas as pd

def clean():
    print("Cleaning OTSO...")
    Delete()
    print("OTSO cleaned")


def Delete():
    script_dir = os.path.dirname(os.path.abspath(__file__))
    setupfile = os.path.join(script_dir, 'setup_complete.txt')

    if os.path.exists(setupfile):
        os.remove(setupfile)

    server_data_folder_path = os.path.join(script_dir, 'Parameters', 'functions', 'ServerData')
    if os.path.exists(server_data_folder_path):
        shutil.rmtree(server_data_folder_path)

def AddStation(Name, Latitude, Longitude):
    script_dir = os.path.dirname(os.path.abspath(__file__))
    stationfile = os.path.join(script_dir, 'Parameters', 'functions', 'StationList.csv')

    df = pd.read_csv(stationfile)

    existing_row = df[df["Name"] == Name]

    if not existing_row.empty:
        print(f"Station '{Name}' already exists:")
        print(existing_row)

        user_input = input("Do you want to overwrite this entry? (y/n): ").strip().lower()
        
        if user_input != "y":
            print("No changes were made.")
            return

        df = df[df["Name"] != Name]

    new_station = pd.DataFrame([[Name, Latitude, Longitude]], columns=["Name", "Latitude", "Longitude"])
    df = pd.concat([df, new_station], ignore_index=True)

    df = df.sort_values(by="Name")

    df.to_csv(stationfile, index=False)

    print(f"Station '{Name}' has been added/updated successfully.")


def RemoveStation(Name):
    script_dir = os.path.dirname(os.path.abspath(__file__))
    stationfile = os.path.join(script_dir, 'Parameters', 'functions', 'StationList.csv')

    df = pd.read_csv(stationfile)

    existing_row = df[df["Name"] == Name]

    if existing_row.empty:
        print(f"Station '{Name}' not found. No changes were made.")
        return

    print(f"Station '{Name}' found:")
    print(existing_row)

    user_input = input("Do you want to delete this entry? (y/n): ").strip().lower()
    
    if user_input != "y":
        print("No changes were made.")
        return

    df = df[df["Name"] != Name]

    df.to_csv(stationfile, index=False)

    print(f"Station '{Name}' has been removed successfully.")


def ListStations():
    script_dir = os.path.dirname(os.path.abspath(__file__))
    stationfile = os.path.join(script_dir, 'Parameters', 'functions', 'StationList.csv')

    df = pd.read_csv(stationfile)

    pd.set_option('display.max_rows', None)
    pd.set_option('display.max_columns', None)

    print(df.to_string(index=False))







import time
from datetime import datetime
import multiprocessing as mp
import os
from . import fortran_calls, readme_generators,cores, misc, cutoff_inputs
import pandas as pd
import sys
import queue

def OTSO_cutoff(Stations,customlocations,startaltitude,cutoff_comp,minaltitude,maxdistance,maxtime,
           serverdata,livedata,vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,Anum,anti,year,
           month,day,hour,minute,second,internalmag,externalmag,
           intmodel,startrigidity,endrigidity,rigiditystep,rigidityscan,
           coordsystem,gyropercent,magnetopause,corenum,azimuth,zenith,g,h, MHDfile, MHDcoordsys,spheresize):

    CutoffInputArray = cutoff_inputs.CutoffInputs(Stations,customlocations,startaltitude,cutoff_comp,minaltitude,maxdistance,maxtime,
           serverdata,livedata,vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,Anum,anti,year,
           month,day,hour,minute,second,internalmag,externalmag,
           intmodel,startrigidity,endrigidity,rigiditystep,rigidityscan,
           coordsystem,gyropercent,magnetopause,corenum,azimuth,zenith,g,h, MHDfile, MHDcoordsys)

    RigidityArray = CutoffInputArray[0]
    DateArray = CutoffInputArray[1]
    Model = CutoffInputArray[2]
    IntModel = CutoffInputArray[3]
    ParticleArray = CutoffInputArray[4]
    IOPT = CutoffInputArray[5]
    WindArray = CutoffInputArray[6]
    Magnetopause = CutoffInputArray[7]
    CoordinateSystem = CutoffInputArray[8]
    MaxStepPercent = CutoffInputArray[9]/100
    EndParams = CutoffInputArray[10]
    Station_Array = CutoffInputArray[11]
    InputtedStations = CutoffInputArray[12]
    Rcomp = CutoffInputArray[13]
    Rscan = CutoffInputArray[14]
    Kp = CutoffInputArray[15]
    corenum = CutoffInputArray[16]
    LiveData = CutoffInputArray[17]
    g = CutoffInputArray[18]
    h = CutoffInputArray[19]

    AntiCheck = ParticleArray[1]

    ChildProcesses = []

    UsedCores = cores.Cores(Station_Array, corenum)
    CoreList = UsedCores.getCoreList()
    Positionlists = UsedCores.getPositions()
    InputtedStations.find_non_matching_stations()

    print("OTSO Cutoff Computation Started")
    start = time.time()
    sys.stdout.write(f"\r{0:.2f}% complete")

# Set the process creation method to 'forkserver'
    try:
        # Check if the start method is already set
        if not mp.get_start_method(allow_none=True):
            mp.set_start_method('spawn')
    except RuntimeError:
        # If the start method is already set, a RuntimeError will be raised
        # You can log or handle this as needed
        pass
# Create a shared message queue for the processes to produce/consume data
    ProcessQueue = mp.Manager().Queue()
    for Data,Core in zip(Positionlists,CoreList):
        Child = mp.Process(target=fortran_calls.fortrancallCutoff,  args=(Data, Core, RigidityArray, DateArray, Model, IntModel, ParticleArray, IOPT,
         WindArray, Magnetopause, CoordinateSystem, MaxStepPercent, EndParams, Rcomp, Rscan, Kp, ProcessQueue, g, h,  MHDfile, MHDcoordsys,spheresize))
        ChildProcesses.append(Child)

    for a in ChildProcesses:
        a.start()

# Wait for child processes to complete

    total_stations = len(Station_Array)
    processed = 0
    results = []

    while processed < total_stations:
      try:
        # Check if the ProcessQueue has any new results
        result_df = ProcessQueue.get(timeout=0.001)  # Use timeout to avoid blocking forever
        results.append(result_df)
        processed += 1

        # Calculate and print the progress
        percent_complete = (processed / total_stations) * 100
        sys.stdout.write(f"\r{percent_complete:.2f}% complete")
        sys.stdout.flush()

      except queue.Empty:
        # Queue is empty, but processes are still running, so we continue checking
        pass
      
      time.sleep(0.0001)

    # Ensure that all processes have completed
    for b in ChildProcesses:
        b.join()


    # Concatenate the results on the index
    merged_df = pd.concat(results, axis=1)
    merged_df = merged_df.sort_index(axis=1)

    print("\nOTSO Cutoff Computation Complete")
    stop = time.time()
    Printtime = round((stop-start),3)
    print("Whole Program Took: " + str(Printtime) + " seconds")
    
    EventDate = datetime(year,month,day,hour,minute,second)
    README = readme_generators.READMECutoff(Station_Array, RigidityArray, EventDate, Model, IntModel, Anum, AntiCheck, IOPT, WindArray, Magnetopause, 
                                            CoordinateSystem, Printtime, MaxStepPercent*100, EndParams, cutoff_comp, Rscan, LiveData, serverdata, kp)
    
    if livedata == 1:
        misc.remove_files()

    return [merged_df, README]


import psutil
import multiprocessing as mp
import numpy as np

class Cores:
   def __init__(self, Positions, CoreNum):
        self.CoreNum = CoreNum
        self.Positions = Positions
        self.CoreNames = []

        self.CheckCoreNum()
        self.CreateCoreList()
        self.SplitPositions()



   def CheckCoreNum(self):
           if(psutil.cpu_count(logical=False) < self.CoreNum):

              print("You have entered an invalid number of cores")
              print("You have " + str(psutil.cpu_count(logical=False)) + " and have tried to use " + str(self.CoreNum) + " cores")
              print("To ensure operational integrity of your computer OTSO will run using 2 less than the max cores available, with a minumum value of 1.")

              self.CoreNum = psutil.cpu_count(logical=False) - 2
              if self.CoreNum <= 0:
                  self.CoreNum = 1

   def CreateCoreList(self):
         i = 1
         while i <= self.CoreNum:
          string = "Core " + str(i)
          self.CoreNames.append(string)
          i += 1

   def SplitPositions(self):
         self.PositionSplit = np.array_split(self.Positions, self.CoreNum)

   def getCoreList(self):
         return self.CoreNames

   def getPositions(self):
         return self.PositionSplit


def trajectory(Stations,rigidity, customlocations=[], startaltitude=20,minaltitude=20,zenith=0,azimuth=0,maxdistance=100,maxtime=0,
           serverdata="OFF",livedata="OFF",vx=-500,vy=0,vz=0,by=5,bz=5,density=1,pdyn=0,Dst=0,
           G1=0,G2=0,G3=0,W1=0,W2=0,W3=0,W4=0,W5=0,W6=0,kp=0,Anum=1,anti="YES",year=2024,
           month=1,day=1,hour=12,minute=0,second=0,internalmag="IGRF",externalmag="TSY89",
           intmodel="Boris",coordsystem="GEO",gyropercent=15,magnetopause="Kobel",corenum=None, g=None, h=None,
           MHDfile=None,MHDcoordsys=None,spheresize=25):
    from .Parameters.functions import otso_trajectory
    import psutil

    if corenum is None:
       corenum = psutil.cpu_count(logical=False) - 2
       if corenum <= 0:
          corenum = 1
    
    arguments = locals()
    for arg in arguments:
       if arguments[arg] is None:
          arguments[arg] = []

    trajectory = otso_trajectory.OTSO_trajectory(Stations,rigidity, customlocations,startaltitude,
           minaltitude,zenith,azimuth,maxdistance,maxtime,serverdata,livedata,vx,vy,vz,by,bz,density,
           pdyn,Dst,G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,Anum,anti,year,
           month,day,hour,minute,second,internalmag,externalmag,intmodel,
           coordsystem,gyropercent,magnetopause,corenum,g,h,MHDfile,MHDcoordsys,spheresize)
    
    return trajectory


import numpy as np
from datetime import datetime,timedelta
import os
from . import date, solar_wind, stations
from . import misc, Request, Server
import warnings # Import warnings

def PlanetInputs(startaltitude,cutoff_comp,minaltitude,maxdistance,maxtime,
           serverdata,livedata,vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,Anum,anti,year,
           month,day,hour,minute,second,internalmag,externalmag,
           intmodel,startrigidity,endrigidity,rigiditystep,rigidityscan,
           gyropercent,magnetopause,corenum,azimuth,zenith, asymptotic,asymlevels,unit,
           latstep,longstep,maxlat,minlat,maxlong,minlong,g,h,MHDfile, MHDcoordsys,
           array_of_lats_and_longs=None,
           grid_params_user_set=False): # Add flag
    
    EventDate = datetime(year,month,day,hour,minute,second)
    DateCreate = date.Date(EventDate)
    DateArray = DateCreate.GetDate()

    if unit not in ["GeV", "GV"]:
         print("Please enter a valid asymlevel unit: ""GeV"" or ""GV"" ")
         exit()

    if anti == "YES":
         AntiCheck = 1
    elif anti == "NO":
         AntiCheck = 0
    else:
         print("Please enter a valid anti value: ""YES"" or ""NO"" ")
         exit()

    if asymptotic not in ["YES","NO"]:
         print("Please enter a valid asymptotic value: ""YES"" or ""NO"" ")
         exit()

    if magnetopause == "Sphere":
         Magnetopause = 0
    elif magnetopause == "aFormisano":
         Magnetopause = 1
    elif magnetopause == "Sibeck":
         Magnetopause = 2
    elif magnetopause == "Kobel":
         Magnetopause = 3
    elif magnetopause == "NONE":
         Magnetopause = 99
    else:
         print("Please enter a valid magnetopause model: ""Sphere"", ""aFormisano"", ""Sibeck"", ""Kobel"", ""NONE"" ")



    if intmodel == "4RK":
         IntModel = 1
    elif intmodel == "Boris":
         IntModel = 2
    elif intmodel == "Vay":
         IntModel = 3
    elif intmodel == "HC":
         IntModel = 4
    else:
         print("Please enter a valid intmodel model: ""4RK"", ""Boris"", ""Vay"", ""HC"" ")
         exit()

    if cutoff_comp == "Vertical":
         CutoffComputation = 0
         Zenith = 0
         Azimuth = 0
    elif cutoff_comp == "Apparent":
         CutoffComputation = 1
         Zenith = 0
         Azimuth = 0
    elif cutoff_comp == "Custom":
         CutoffComputation = 2
         Zenith = zenith
         Azimuth = azimuth
    else:
         print("Please enter a valid cutoff_comp: ""Vertical"", ""Apparent"", or ""Custom"" ")
         exit()

    if serverdata == "ON":
         ServerData = 1
    elif serverdata == "OFF":
         ServerData = 0
    else:
         print("Please enter a valid serverdata value: ""ON"" or ""OFF"" ")
         exit()

    if livedata == "ON":
         LiveData = 1
    elif livedata == "OFF":
         LiveData = 0
    else:
         print("Please enter a valid livedata value: ""ON"" or ""OFF"" ")
         exit()
    
    if internalmag == "NONE":
         Internal = 0
         if not g or not h: 
            g = [0] * 105
            h = [0] * 105
    elif internalmag == "IGRF":
         Internal = 1
         if not g or not h: 
            g = [0] * 105
            h = [0] * 105
    elif internalmag == "Dipole":
         Internal = 2
         if not g or not h: 
            g = [0] * 105
            h = [0] * 105
    elif internalmag == "Custom Gauss":
         Internal = 4
         if not g or not h:
              print("Please enter values for the g and h Gaussian coefficients to use the Custom Gauss option")
              exit()
         elif len(g) != 105:
              print(f"There should be 105 g coefficents in the inputted list, you have entered {len(g)}")
              exit()
         elif len(h) != 105:
              print(f"There should be 105 h coefficents in the inputted list, you have enetered {len(h)}")
    else:
         print("Please enter a valid internalmag model: ""NONE"",""IGRF"",""Dipole"", or ""Custom Gauss""")
         exit()
      
    if externalmag == "NONE":
         External = 0
    elif externalmag == "TSY87short":
         External = 1
    elif externalmag == "TSY87long":
         External = 2
    elif externalmag == "TSY89":
         External = 3
    elif externalmag == "TSY96":
         External = 4
    elif externalmag == "TSY01":
         External = 5
    elif externalmag == "TSY01S":
         External = 6
    elif externalmag == "TSY04":
         External = 7
    elif externalmag == "TSY89_BOBERG":
         External = 8
    elif externalmag == "MHD":
         External = 99
         if not os.path.exists(MHDfile):
            print(f"The file '{MHDfile}' does not exist.")
            exit()
    else:
         print("Please enter a valid externalmag model: ""NONE"", ""TSY87short"",""TSy87long"",""TSY89"",""TSY89_BOBERG"",""TSY96"",""TSY01"",""TSY01S"",""TSY04""")
         exit()

    
    if rigidityscan == "ON":
         Rscan = 1
    elif rigidityscan == "OFF":
         Rscan = 0
    else:
         print("Please enter a valid rigidityscan value: ""ON"" or ""OFF"" ")
         exit()


    misc.DataCheck(ServerData,LiveData,EventDate)

    IOPTinput = misc.IOPTprocess(kp)
    KpS = 0

    if ServerData == 1:
         if int(EventDate.year) >= 1981:
              Server.DownloadServerFile(int(EventDate.year))
         elif int(EventDate.year) < 1981 and int(EventDate.year) > 1963:
              Server.DownloadServerFileLowRes(int(EventDate.year))
         else:
              print("Server data only valid for 1963 to present, please enter a valid date.")
         ByS, BzS, VS, DensityS, PdynS, KpS, DstS, G1S, G2S, G3S, W1S, W2S, W3S, W4S, W5S, W6S = Server.GetServerData(EventDate,External)
         IOPTinput = misc.IOPTprocess(KpS)
         WindCreate = solar_wind.Solar_Wind(VS, vy, vz, ByS, BzS, DensityS, PdynS, DstS, G1S, G2S, G3S, W1S, W2S, W3S, W4S, W5S, W6S, KpS)
         WindArray = WindCreate.GetWind()
         
    if LiveData == 1:
         misc.DateCheck(EventDate)
         DstLive, VxLive, DensityLive, ByLive, BzLive, IOPTLive, G1Live, G2Live, G3Live, KpLive = Request.Get_Data(EventDate)
         PdynLive = misc.Pdyn_comp(DensityLive,VxLive)
         IOPTinput = IOPTLive
         WindCreate = solar_wind.Solar_Wind(VxLive, vy, vz, ByLive, BzLive, DensityLive, PdynLive, DstLive, G1Live, G2Live, G3Live, W1, W2, W3, W4, W5, W6, KpLive)
         WindArray = WindCreate.GetWind()

    if ServerData == 0 and LiveData == 0:
          if vx > 0:
               vx = -1*vx
          WindCreate = solar_wind.Solar_Wind(vx, vy, vz, by, bz, density, pdyn, Dst, G1, G2, G3, W1, W2, W3, W4, W5, W6, kp)
          WindArray = WindCreate.GetWind()

    RigidityArray = [startrigidity,endrigidity,rigiditystep]

    MagFieldModel = np.array([Internal,External])

    EndParams = [minaltitude,maxdistance,maxtime]

    ParticleArray = [Anum,AntiCheck]

    misc.ParamCheck(startaltitude,year,Internal,EndParams)

    # --- Start: Coordinate Generation Logic --- 
    coordinate_pairs = []
    LatitudeList_meta = [] # For metadata/README
    LongitudeList_meta = [] # For metadata/README

    # Check and handle coordinate input
    if array_of_lats_and_longs is not None:
        try:
            coord_array = np.array(array_of_lats_and_longs)
            if coord_array.ndim != 2 or coord_array.shape[1] != 2:
                raise ValueError("array_of_lats_and_longs must be a list of pairs or a Nx2 array.")
            if not np.issubdtype(coord_array.dtype, np.number):
                 raise ValueError("Coordinates in array_of_lats_and_longs must be numeric.")
            coordinate_pairs = coord_array.tolist()
        except Exception as e:
            raise ValueError(f"Error processing array_of_lats_and_longs: {e}")

        if not coordinate_pairs:
             raise ValueError("Provided array_of_lats_and_longs is empty.")

        lats = coord_array[:, 0]
        lons = coord_array[:, 1]
        LatitudeList_meta = sorted(np.unique(lats), reverse=True)
        LongitudeList_meta = sorted(np.unique(lons))

        # --- Conditional Warning --- 
        # Only warn if grid params were explicitly set by the user
        if grid_params_user_set:
        # if latstep is not None or longstep is not None or maxlat is not None or minlat is not None or maxlong is not None or minlong is not None: # Old check
            warnings.warn("Both array_of_lats_and_longs and step/range parameters were provided. The array_of_lats_and_longs will be used.", UserWarning)
        # --- End Conditional Warning ---
    else:
        # Generate grid using step/range parameters (which now have defaults)
        # Remove the check for missing parameters, as they have defaults now.
        # required_params = [latstep, longstep, maxlat, minlat, maxlong, minlong]
        # if any(p is None for p in required_params):
        #      raise ValueError("Missing latitude/longitude step or range parameters when array_of_lats_and_longs is not provided.")
        
        # Keep the check for zero step
        if latstep == 0 or longstep == 0:
            raise ValueError("latstep and longstep must be non-zero for grid generation.")

        if latstep > 0:
             warnings.warn("latstep is positive. Latitude generation usually expects a negative step to go from maxlat down to minlat. Proceeding, but check your parameters.", UserWarning)

        lat_start, lat_end = (maxlat, minlat) if latstep < 0 else (minlat, maxlat)
        lon_start, lon_end = (minlong, maxlong)

        epsilon_lat = abs(latstep / 1000.0)
        epsilon_lon = abs(longstep / 1000.0)
        
        _LatitudeList_np = np.arange(lat_start, lat_end + (np.sign(latstep) * epsilon_lat), latstep)
        _LongitudeList_np = np.arange(lon_start, lon_end + epsilon_lon, longstep)

        if _LatitudeList_np.size == 0:
            raise ValueError(f"Generated LatitudeList is empty. Check maxlat ({maxlat}), minlat ({minlat}), and latstep ({latstep}).")
        if _LongitudeList_np.size == 0:
            raise ValueError(f"Generated LongitudeList is empty. Check maxlong ({maxlong}), minlong ({minlong}), and longstep ({longstep}).")

        LatitudeList_meta = _LatitudeList_np.tolist()
        LongitudeList_meta = _LongitudeList_np.tolist()
        coordinate_pairs = [[lat, lon] for lat in LatitudeList_meta for lon in LongitudeList_meta]
    # --- End: Coordinate Generation Logic --- 

    PlanetInputArray = [coordinate_pairs, RigidityArray, DateArray, MagFieldModel, IntModel, ParticleArray, IOPTinput, WindArray, Magnetopause, gyropercent, EndParams, CutoffComputation, Rscan, Zenith, Azimuth, corenum, asymptotic, asymlevels, startaltitude, LiveData, AntiCheck, g, h, LatitudeList_meta, LongitudeList_meta]

    return PlanetInputArray


import os
import psutil
import pandas as pd
import glob
import multiprocessing as mp
import shutil
from datetime import datetime,timedelta

def ParamCheck(Alt, Year, Internal, EndParams):
       if EndParams[0] > Alt:
        print("ERROR: Inputted minimum altitude is larger than starting altitude. Value must be less than or equal to the starting altitude. Please check inputs. \nOTSO program will now terminate.")
        exit()

       if Internal == 1:
         if Year<1900 or Year>2030:
          print("ERROR: IGRF 14 model currently only works for dates between 1900 to 2030. Please select a date within this range to use the IGRF model")
          exit()

def DataCheck(ServerData, LiveData, EventDate):
       current_date = datetime.utcnow()
       if (ServerData == 1 or LiveData == 1) and EventDate > current_date:
        print("ERROR: Future date entered. No valid data available. \nOTSO program will now terminate.")
        exit()
       #if ServerData == 1 and EventDate.year < 1981:
        #print("ERROR: OMNI data only goes back to 1981. No valid data for entered date. \nOTSO program will now terminate.")
        #exit()
       if ServerData == 1 and LiveData == 1:
        print("ERROR: You have requested live and server data. Only one can be selected. \nOTSO program will now terminate.")
        exit()





def PlanetFile(final_directory):
 
 planetfile = "Planet.csv"
 Planetfile_path = os.path.join(final_directory, planetfile)
 if os.path.isfile(Planetfile_path):
    os.remove(Planetfile_path)

 files = os.path.join(final_directory, "*.csv")
 files = glob.glob(files)
 df = pd.concat(map(pd.read_csv, files), ignore_index=True)
 df_sorted = df.sort_values(by=[df.columns[0], df.columns[1]])
 os.makedirs(final_directory, exist_ok=True)  
 df_sorted.to_csv(final_directory + '/Planet.csv', index=False)

 for i in files:
   os.remove(i)

def FlightFile(final_directory):
 
 planetfile = "Flight.csv"
 Planetfile_path = os.path.join(final_directory, planetfile)
 if os.path.isfile(Planetfile_path):
    os.remove(Planetfile_path)

 files = glob.glob(os.path.join(final_directory, "*.csv"))
 files = [f for f in files if not f.endswith("Flight_Params.csv")]
 df = pd.concat(map(pd.read_csv, files), ignore_index=True)
 df['Time'] = pd.to_datetime(df['Time'], format='%d-%m-%Y %H-%M-%S')
 df_sorted = df.sort_values(by=[df.columns[0]])
 os.makedirs(final_directory, exist_ok=True)  
 df_sorted.to_csv(final_directory + '/Flight.csv', index=False)

 for i in files:
   os.remove(i)

def FlightCopy(final_directory):
 
 current_directory = os.getcwd()
 planetfile = "Parameters\\Flight_Params.csv"
 Planetfile_path = os.path.join(current_directory, planetfile)
 
 shutil.copy2(Planetfile_path, final_directory)

def CheckCoreNumPlanet(x):
  NewCore = x
  if(psutil.cpu_count(logical=False) < x):
    print("ERROR: You have entered an invalid number of cores")
    print("You have " + str(psutil.cpu_count(logical=False)) + " and have tried to use " + str(x) + " cores")
    print("To ensure operational integrity of your computer OTSO will run using 2 less than the max cores available, with a minumum value of 1.")
    NewCore = psutil.cpu_count(logical=False) - 2
    if NewCore <= 0:
      NewCore = 1
  return NewCore

def day_of_year_to_date(doy, year):
    # Define days in each month for both leap and non-leap years
    days_in_months_non_leap = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    days_in_months_leap = [31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

    # Check if the year is a leap year
    if (year % 4 == 0 and year % 100 != 0) or year % 400 == 0:
        days_in_months = days_in_months_leap
    else:
        days_in_months = days_in_months_non_leap

    # Initialize variables to accumulate days and find the month
    accumulated_days = 0
    month = 0

    # Find the month and day corresponding to the day of the year (DOY)
    for i in range(12):
        accumulated_days += days_in_months[i]
        if doy <= accumulated_days:
            month = i + 1
            day = doy - (accumulated_days - days_in_months[i])
            break

    # Return the month and day as a tuple
    return (month, day)

def DateCheck(Date):
     current_date = datetime.utcnow()
     seven_days_ago = current_date - timedelta(days=7)

     if Date < seven_days_ago:
         print("ERROR: Inputed date is over 7 days ago from current time (" + str(current_date) + ").\n live data only available for the last week. \nOTSO will now terminate.")
         exit()

     return

def Pdyn_comp(Density,Vx):
    Pressure = ((((Density))*10**6)*(1.672621898e-27)*((Vx*1000)**2))*10**(9)
    Pressure = round(Pressure,3)
    return Pressure

def remove_folder(folder_path):
    try:
        shutil.rmtree(folder_path)
        #print("Folder deleted successfully.")
    except FileNotFoundError:
        print("The folder does not exist.")
    except PermissionError:
        print("You do not have permission to delete this folder.")
    except Exception as e:
        print(f"Error: {e}")

def remove_files():
    file_list = ["Dst_data.txt","Kp_data.txt","Magnetic_data.csv","Magnetic_data.json","space_data.csv","space_data.json"]
    directory1 = os.path.dirname(os.path.dirname(__file__))
    directory = os.path.join(directory1, "functions")
    for file_name in file_list:
        file_path = os.path.join(directory, file_name)
        try:
            os.remove(file_path)
        except FileNotFoundError:
            print(f"File not found: {file_path}")
        except PermissionError:
            print(f"Permission denied: {file_path}")
        except Exception as e:
            print(f"Error deleting {file_path}: {e}")


def IOPTprocess(Kp):
    if Kp >= 6:
        IOPT = 7
    else:
        IOPT = Kp + 1

    return IOPT


import time
from datetime import datetime, timedelta, date
import numpy as np
import calendar

class Date:
 def __init__(self, Date):
    self.year = Date.year
    self.hour = Date.hour
    self.minute = Date.minute
    self.secs = Date.second
    self.seconds = timedelta(hours = Date.hour, minutes = Date.minute, seconds = Date.second).total_seconds()
    self.day = Date.timetuple().tm_yday

    self.DateArray = np.array([self.year, self.day, self.hour, self.minute, self.secs, self.seconds])

 def GetDate(self):
    return self.DateArray
 
def convert_to_datetime(date_list):
    year, day_of_year, hour, minute, second, seconds = date_list

    year = int(year)
    day_of_year = int(day_of_year)
    hour = int(hour)
    minute = int(minute)
    second = int(second)
    
    # Automatically determine if it's a leap year
    is_leap_year = calendar.isleap(year)
    
    # Convert day_of_year to month and day
    month, day = day_of_year_to_date(day_of_year, is_leap_year)
    
    # Create the date from the year, month, and day
    date = datetime(year, month, day)
    
    # Create a timedelta for the time portion
    time_delta = timedelta(hours=hour, minutes=minute, seconds=second)
    
    # Add the time to the date
    full_datetime = date + time_delta
    return full_datetime

# Function to convert day_of_year to month and day
def day_of_year_to_date(day_of_year, is_leap_year=False):
    months_days = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    
    if is_leap_year:
        months_days[1] = 29
    
    month = 0
    while day_of_year > months_days[month]:
        day_of_year -= months_days[month]
        month += 1
    
    return month + 1, day_of_year


def planet(startaltitude=20,cutoff_comp="Vertical",minaltitude=20,maxdistance=100,maxtime=0,
           serverdata="OFF",livedata="OFF",vx=-500,vy=0,vz=0,by=5,bz=5,density=1,pdyn=0,Dst=0,
           G1=0,G2=0,G3=0,W1=0,W2=0,W3=0,W4=0,W5=0,W6=0,kp=0,anti="YES",year=2024,
           month=1,day=1,hour=12,minute=0,second=0,internalmag="IGRF",externalmag="TSY89",
           intmodel="Boris",startrigidity=20,endrigidity=0,rigiditystep=0.01,rigidityscan="ON",
           gyropercent=15,magnetopause="Kobel",corenum=None, azimuth=0, zenith=0, asymptotic="NO",
           asymlevels = [0.1,0.3,0.5,1,2,3,4,5,6,7,8,9,10,15,20,30,50,70,100,300,500,700,1000],unit="GeV",
           latstep=-5,longstep=5, maxlat=90,minlat=-90,maxlong=360,minlong=0,g=None,h=None,MHDfile=None,MHDcoordsys=None,
           array_of_lats_and_longs=None,spheresize=25):
    from .Parameters.functions import otso_planet
    import psutil

    if corenum is None:
       corenum = psutil.cpu_count(logical=False) - 2
       if corenum <= 0:
          corenum = 1
    
    # Capture all local variables, including defaults and user-provided kwargs
    arguments = locals()
    
    # --- Check if grid parameters were explicitly set by the user --- 
    grid_param_keys = {'latstep', 'longstep', 'maxlat', 'minlat', 'maxlong', 'minlong'}
    # We need a way to know what was *actually* passed in the function call.
    # inspect.signature might be too complex here. A common pattern is to use a 
    # sentinel default value (like object()) to distinguish, but that requires changing defaults.
    # Alternative: Check if the values in `arguments` differ from the defaults defined here.
    # This assumes defaults are static, which they are in this case.
    defaults = {
        'latstep': -5, 'longstep': 5, 'maxlat': 90, 'minlat': -90, 'maxlong': 360, 'minlong': 0
    }
    grid_params_user_set = False
    for key in grid_param_keys:
        if arguments[key] != defaults[key]:
            grid_params_user_set = True
            break
    # --- End Check --- 

    # Exclude specific keys where None is a valid/handled input
    excluded_keys = {'g', 'h', 'array_of_lats_and_longs'}
    for arg, value in arguments.items():
       if arg not in excluded_keys and value is None:
          # Defaulting None to [] might still be needed for some args?
          arguments[arg] = [] 

    # Pass the flag indicating user-set grid params
    planet_result = otso_planet.OTSO_planet(
        startaltitude, cutoff_comp, minaltitude, maxdistance, maxtime,
        serverdata, livedata, vx, vy, vz, by, bz, density, pdyn, Dst,
        G1, G2, G3, W1, W2, W3, W4, W5, W6, kp, anti, year,
        month, day, hour, minute, second, internalmag, externalmag,
        intmodel, startrigidity, endrigidity, rigiditystep, rigidityscan,
        gyropercent, magnetopause, corenum, azimuth, zenith, asymptotic, asymlevels, unit,
        latstep, longstep, maxlat, minlat, maxlong, minlong, g, h, MHDfile, MHDcoordsys,spheresize,
        array_of_lats_and_longs=array_of_lats_and_longs,
        grid_params_user_set=grid_params_user_set
    )
    
    return planet_result


import numpy as np

class Solar_Wind:
 def __init__(self, Vx, Vy, Vz, By, Bz, Density, Pdyn, Dst, G1, G2, G3, W1, W2, W3, W4, W5, W6, Kp):
    self.Vx = Vx
    self.Vy = Vy
    self.Vz = Vz
    self.By = By
    self.Bz = Bz
    self.Density = Density
    self.Pdyn = Pdyn
    self.Dst = Dst
    self.G1 = G1
    self.G2 = G2
    self.G3 = G3
    self.W1 = W1
    self.W2 = W2
    self.W3 = W3
    self.W4 = W4
    self.W5 = W5
    self.W6 = W6
    self.KpIndex = Kp

    self.DstBob = 0

    if Kp > 6:
      self.DstBob = (Kp-6)*(-50)

    if Dst != 0:
      self.DstBob = Dst

    self.WindArray = np.array([self.Vx, self.Vy, self.Vz, self.By, self.Bz, self.Density, self.Pdyn, self.Dst, self.G1, self.G2, self.G3, self.W1, self.W2, self.W3, self.W4, self.W5, self.W6, self.KpIndex, self.DstBob])

 def GetWind(self):
    return self.WindArray


import numpy as np
import pandas as pd
import os

class Stations:
 def __init__(self, Stations, Altitude, Zenith, Azimuth):
    self.StationList = np.array([])
    self.Stations = Stations
    self.Altitude = Altitude
    self.Zenith = Zenith
    self.Azimuth = Azimuth
   
    script_dir = os.path.dirname(os.path.realpath(__file__))
    subfolder_name = ''
    csv_file_name = 'StationList.csv'

    csv_file_path = os.path.join(script_dir, subfolder_name, csv_file_name)
    data = pd.read_csv(csv_file_path)
    data['Altitude'] = self.Altitude
    data['Zenith'] = self.Zenith
    data['Azimuth'] = self.Azimuth
    
    self.df_stations = data

    self.CreateList(data)


 def AddLocation(self, CustomList):
    
    for x in CustomList:
       NewStation = np.array([x[0], x[1], x[2], self.Altitude, self.Zenith, self.Azimuth])
       self.StationList = np.vstack((self.StationList, NewStation))

 def AddLocationFlight(self, CustomList, alt):
    
    for x in CustomList:
       NewStation = np.array([x[0], x[1], x[2], alt, self.Zenith, self.Azimuth])
       self.StationList = np.vstack((self.StationList, NewStation))

 def CreateList(self, data):
    NewList = data[data["Name"].isin(self.Stations)]
    self.StationList = NewList.to_numpy()

 def GetStations(self):
    return self.StationList
 
 def find_non_matching_stations(self):
    csv_stations = self.df_stations['Name'].tolist()
    non_matching_stations = [name for name in self.Stations if name not in csv_stations]
    
    if len(non_matching_stations) > 0:
      print("ERROR: The following stations were not found in the StationList.csv file:")
      for x in non_matching_stations:
         print(x)
      print("Computations for these stations could not be conducted\n\
      To Fix:\n\
            Double check spelling\n\
            Add the station to the StationList.csv within the Parameters/functions folder\n\
            Manually add the station using customlocations folder\n", sep='\n') 
      


from cmath import atan
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import cartopy.crs as ccrs
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
import matplotlib.ticker as mticker
import math

#data = pd.read_csv("GLE66_TSY01_DATA.csv")

#IMFy = data["IMFy"]
#IMFz = data["IMFz"]
#Speed = data["Speed"]

#################################################################
# Control Test
# G1 = 5.55
# G2 = 10
IMFy = [-2.83]
IMFz = [0.48]
Speed = [648.2]
Density = [1.47]
#############################################################

G1 = 0
G2 = 0
G3 = 0

for (By, Bz, V, N) in zip(IMFy, IMFz, Speed, Density):
    By = By
    Bz = Bz
    V = V
    W = 1/len(IMFy)


    B = (By*By + Bz*Bz)**(0.5)
    h = (((B/40)**(2))/(1 + B/40))

    if(By == 0 and Bz == 0):
        phi = 0
    else:
        phi = math.atan2(By,Bz)
        if(phi <= 0):
            phi = phi + 2*math.pi

    if(Bz < 0):
        Bs = abs(Bz)
    elif Bz >= 0:
        Bs = 0
    
    G1 = G1 + (W)*V*h*(math.sin(phi/2)**3)
    G2 = G2 + (0.005)*(W)*(V)*Bs
    G3 = G3 + (N*V*Bs)/2000
    
print("G1 = " + str(G1))
print("G2 = " + str(G2))
print("G3 = " + str(G3))

    


import time
from datetime import datetime
import multiprocessing as mp
import os
from . import fortran_calls, readme_generators,cores, misc, trajectory_inputs
import pandas as pd
import sys
import queue
import random
import numpy as np

def OTSO_trajectory(Stations,rigidity, customlocations,startaltitude,
           minaltitude,zenith,azimuth,maxdistance,maxtime,serverdata,livedata,vx,vy,vz,by,bz,density,
           pdyn,Dst,G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,Anum,anti,year,
           month,day,hour,minute,second,internalmag,externalmag,intmodel,
           coordsystem,gyropercent,magnetopause,corenum,g,h,MHDfile,MHDcoordsys,spheresize):

    TrajectoryInputArray = trajectory_inputs.TrajectoryInputs(Stations,rigidity, customlocations,startaltitude,
           minaltitude,zenith,azimuth,maxdistance,maxtime,serverdata,livedata,vx,vy,vz,by,bz,density,
           pdyn,Dst,G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,Anum,anti,year,
           month,day,hour,minute,second,internalmag,externalmag,intmodel,
           coordsystem,gyropercent,magnetopause,corenum,g,h,MHDfile,MHDcoordsys)

    Rigidity = TrajectoryInputArray[0]
    DateArray = TrajectoryInputArray[1]
    Model = TrajectoryInputArray[2]
    IntModel = TrajectoryInputArray[3]
    ParticleArray = TrajectoryInputArray[4]
    IOPT = TrajectoryInputArray[5]
    WindArray = TrajectoryInputArray[6]
    Magnetopause = TrajectoryInputArray[7]
    CoordinateSystem = TrajectoryInputArray[8]
    MaxStepPercent = TrajectoryInputArray[9]/100
    EndParams = TrajectoryInputArray[10]
    Station_Array = TrajectoryInputArray[11]
    InputtedStations = TrajectoryInputArray[12]
    Kp = TrajectoryInputArray[13]
    CoreNum = TrajectoryInputArray[14]
    LiveData = TrajectoryInputArray[15]
    ServerData = TrajectoryInputArray[16]
    g = TrajectoryInputArray[17]
    h = TrajectoryInputArray[18]


    ChildProcesses = []

    UsedCores = cores.Cores(Station_Array, CoreNum)
    CoreList = UsedCores.getCoreList()
    Positionlists = UsedCores.getPositions()

    start = time.time()
    InputtedStations.find_non_matching_stations()
    print("OTSO Trajectory Computation Started")
    sys.stdout.write(f"\r{0:.2f}% complete")

# Set the process creation method to 'forkserver'
    try:
        # Check if the start method is already set
        if not mp.get_start_method(allow_none=True):
            mp.set_start_method('spawn')
    except RuntimeError:
        # If the start method is already set, a RuntimeError will be raised
        # You can log or handle this as needed
        pass
# Create a shared message queue for the processes to produce/consume data
    ProcessQueue = mp.Manager().Queue()
    for Data,Core in zip(Positionlists,CoreList):
        Child = mp.Process(target=fortran_calls.fortrancallTrajectory,  args=(Data, Core, Rigidity, DateArray, Model, IntModel, ParticleArray, IOPT, 
                                                                              WindArray, Magnetopause, CoordinateSystem, MaxStepPercent, EndParams, 
                                                                              ProcessQueue,g,h, MHDfile, MHDcoordsys,spheresize))
        ChildProcesses.append(Child)

    for a in ChildProcesses:
        a.start()

    total_stations = len(Station_Array)
    processed = 0
    results = []

    while processed < total_stations:
      try:
        # Check if the ProcessQueue has any new results
        result_df = ProcessQueue.get(timeout=0.001)  # Use timeout to avoid blocking forever
        results.append(result_df)
        processed += 1
  
        # Calculate and print the progress
        percent_complete = (processed / total_stations) * 100
        sys.stdout.write(f"\r{percent_complete:.2f}% complete")
        sys.stdout.flush()

      except queue.Empty:
        # Queue is empty, but processes are still running, so we continue checking
        pass
      
      time.sleep(0.0001)

    for b in ChildProcesses:
        b.join()

# Wait for child processes to complete
    combined_dict = {}
    for d in results:
        combined_dict.update(d)

    print("\nOTSO Trajectory Computation Complete")
    stop = time.time()
    Printtime = round((stop-start),3)
    print("Whole Program Took: " + str(Printtime) + " seconds")

    EventDate = datetime(year,month,day,hour,minute,second)
    AntiCheck = ParticleArray[0]
    Anum = ParticleArray[1]
    readme = readme_generators.READMETrajectory(Station_Array, Rigidity, EventDate, Model, IntModel, Anum, AntiCheck, IOPT, WindArray, Magnetopause, CoordinateSystem, Printtime, MaxStepPercent*100, EndParams, LiveData, Kp, serverdata)

    if LiveData == 1:
        misc.remove_files()

    return [combined_dict, readme]


import time
from datetime import datetime
import multiprocessing as mp
import os
from . import fortran_calls, readme_generators,cores, misc, cone_inputs
import pandas as pd
import sys
import queue

def OTSO_cone(Stations,customlocations,startaltitude,minaltitude,zenith,azimuth,maxdistance,maxtime,
           serverdata,livedata,vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,Anum,anti,year,
           month,day,hour,minute,second,internalmag,externalmag,
           intmodel,startrigidity,endrigidity,rigiditystep,
           coordsystem,gyropercent,magnetopause,corenum,g,h,MHDfile,MHDcoordsys,spheresize):

    ConeInputArray = cone_inputs.ConeInputs(Stations,customlocations,startaltitude,minaltitude,zenith,azimuth,maxdistance,maxtime,
           serverdata,livedata,vx,vy,vz,by,bz,density,pdyn,Dst,
           G1,G2,G3,W1,W2,W3,W4,W5,W6,kp,Anum,anti,year,
           month,day,hour,minute,second,internalmag,externalmag,
           intmodel,startrigidity,endrigidity,rigiditystep,
           coordsystem,gyropercent,magnetopause,corenum,g,h,MHDfile,MHDcoordsys)

    RigidityArray = ConeInputArray[0]
    DateArray = ConeInputArray[1]
    Model = ConeInputArray[2]
    IntModel = ConeInputArray[3]
    ParticleArray = ConeInputArray[4]
    IOPT = ConeInputArray[5]
    WindArray = ConeInputArray[6]
    Magnetopause = ConeInputArray[7]
    CoordinateSystem = ConeInputArray[8]
    MaxStepPercent = ConeInputArray[9]/100
    EndParams = ConeInputArray[10]
    Station_Array = ConeInputArray[11]
    InputtedStations = ConeInputArray[12]
    Kp = ConeInputArray[13]
    CoreNum = ConeInputArray[14]
    LiveData = ConeInputArray[15]
    g = ConeInputArray[16]
    h = ConeInputArray[17]

    AntiCheck = ParticleArray[1]

    ChildProcesses = []
    results = []

    UsedCores = cores.Cores(Station_Array, CoreNum)
    CoreList = UsedCores.getCoreList()
    Positionlists = UsedCores.getPositions()

    start = time.time()
    InputtedStations.find_non_matching_stations()
    print("OTSO Cone Computation Started")
    sys.stdout.write(f"\r{0:.2f}% complete")


    try:
        if not mp.get_start_method(allow_none=True):
            mp.set_start_method('spawn')
    except RuntimeError:

        pass
# Create a shared message queue for the processes to produce/consume data
    ProcessQueue = mp.Manager().Queue()
    for Data,Core in zip(Positionlists,CoreList):
        Child = mp.Process(target=fortran_calls.fortrancallCone,  args=(Data, Core, RigidityArray, DateArray, Model, IntModel, ParticleArray, IOPT, WindArray, 
                                                                        Magnetopause, CoordinateSystem, MaxStepPercent, EndParams, ProcessQueue,g,h,MHDfile, MHDcoordsys,
                                                                        spheresize))
        ChildProcesses.append(Child)

    for a in ChildProcesses:
        a.start()

# Wait for child processes to complete

    results = []
    total_stations = len(Station_Array)
    processed = 0

    while processed < total_stations:
      try:
        # Check if the ProcessQueue has any new results
        result_df = ProcessQueue.get(timeout=0.001)  # Use timeout to avoid blocking forever
        results.append(result_df)
        processed += 1

        # Calculate and print the progress
        percent_complete = (processed / total_stations) * 100
        sys.stdout.write(f"\r{percent_complete:.2f}% complete")
        sys.stdout.flush()

      except queue.Empty:
        # Queue is empty, but processes are still running, so we continue checking
        pass
      
      time.sleep(0.0001)

    # Ensure that all processes have completed
    for b in ChildProcesses:
        b.join()

    conedf_list = []
    Rigiditylist = []

    for x in results:
        conedf_list.append(x[0])
        Rigiditylist.append(x[1])

    merged_cone_df = conedf_list[0]
    for df in conedf_list[1:]:
        merged_cone_df = pd.merge(merged_cone_df, df, on='R [GV]')
    cols = ['R [GV]'] + [col for col in merged_cone_df.columns if col != 'R [GV]']
    merged_cone_df = merged_cone_df[cols]

    merged_cone_df = merged_cone_df[['R [GV]'] + sorted(merged_cone_df.columns.drop('R [GV]'))]

    merged_R_df = pd.concat(Rigiditylist, axis=1)
    merged_R_df = merged_R_df.sort_index(axis=1)

    print("\nOTSO Cone Computation Complete")
    stop = time.time()
    Printtime = round((stop-start),3)
    print("Whole Program Took: " + str(Printtime) + " seconds")
    
    EventDate = datetime(year,month,day,hour,minute,second)
    README = readme_generators.READMECone(Station_Array, RigidityArray, EventDate, Model, IntModel, Anum, AntiCheck, IOPT, WindArray, Magnetopause, 
                                          CoordinateSystem, Printtime, MaxStepPercent*100, EndParams, LiveData, serverdata, kp)

    if LiveData == 1:
        misc.remove_files()
    
    return [merged_cone_df, merged_R_df, README]